"""
A module for representing simple graph-oriented views of an ontology

See also:

 - ontol_factory.py

"""

import networkx as nx
import logging
import re

logger = logging.getLogger(__name__)

class Ontology():
    """An object that represents a basic graph-oriented view over an ontology.

    The ontology may be represented in memory, or it may be located
    remotely. See subclasses for details.

    The default implementation is an in-memory wrapper onto the python networkx library

    """

    def __init__(self,
                 handle=None,
                 id=None,
                 graph=None,
                 xref_graph=None,
                 meta=None,
                 payload=None,
                 graphdoc=None):
        """
        initializes based on an ontology name.

        **Note**: do not call this directly, use OntologyFactory instead
        """
        self.handle = handle
        self.meta = meta
        if id is None:
            if payload is not None:
                id = payload.get('id')
            if id is None:
                id = handle
        self.id = id

        # networkx object
        self.graph = graph
        if self.graph is None:
            self.graph = nx.MultiDiGraph()
        logger.debug('Graph initialized, nodes={}'.format(self.graph.nodes()))
        self.xref_graph = xref_graph

        # obograph
        self.graphdoc = graphdoc

        self.all_logical_definitions = []
        self.all_property_chain_axioms = []

        # alternatively accept a payload object
        if payload is not None:
            self.meta = payload.get('meta')
            self.graph = payload.get('graph')
            self.xref_graph = payload.get('xref_graph')
            self.graphdoc = payload.get('graphdoc')
            self.all_logical_definitions = payload.get('logical_definitions')
            self.all_property_chain_axioms = payload.get('property_chain_axioms')

    def __str__(self):
        return '{} handle: {} meta: {}'.format(self.id, self.handle, self.meta)
    def __repr__(self):
        return self.__str__()

    def get_graph(self):
        """
        Return a networkx graph for the whole ontology.

        Note: Only implemented for *eager* implementations

        Return
        ------
        nx.MultiDiGraph
            A networkx MultiDiGraph object representing the complete ontology
        """
        return self.graph

    # consider caching
    def get_filtered_graph(self, relations=None, prefix=None):
        """
        Returns a networkx graph for the whole ontology, for a subset of relations

        Only implemented for eager methods.

        Implementation notes: currently this is not cached

        Arguments
        ---------
          - relations : list
             list of object property IDs, e.g. subClassOf, BFO:0000050. If empty, uses all.
          - prefix : String
             if specified, create a subgraph using only classes with this prefix, e.g. ENVO, PATO, GO
        Return
        ------
        nx.MultiDiGraph
            A networkx MultiDiGraph object representing the filtered ontology
        """

        # trigger synonym cache
        self.all_synonyms()
        self.all_obsoletes()

        # default method - wrap get_graph
        srcg = self.get_graph()
        if prefix is not None:
            srcg = srcg.subgraph([n for n in srcg.nodes() if n.startswith(prefix+":")])
        if relations is None:
            logger.info("No filtering on "+str(self))
            return srcg
        logger.info("Filtering {} for {}".format(self, relations))
        g = nx.MultiDiGraph()

        # TODO: copy full metadata
        logger.info("copying nodes")
        for (n,d) in srcg.nodes(data=True):
            g.add_node(n, **d)

        logger.info("copying edges")
        num_edges = 0
        for (x,y,d) in srcg.edges(data=True):
            if d['pred'] in relations:
                num_edges += 1
                g.add_edge(x,y,**d)
        logger.info("Filtered edges: {}".format(num_edges))
        return g

    def merge(self, ontologies):
        """
        Merges specified ontology into current ontology
        """
        if self.xref_graph is None:
            self.xref_graph = nx.MultiGraph()
        logger.info("Merging source: {} xrefs: {}".format(self, len(self.xref_graph.edges())))
        for ont in ontologies:
            logger.info("Merging {} into {}".format(ont, self))
            g = self.get_graph()
            srcg = ont.get_graph()
            for n in srcg.nodes():
                g.add_node(n, **srcg.nodes[n])
            for (o,s,m) in srcg.edges(data=True):
                g.add_edge(o,s,**m)
            if ont.xref_graph is not None:
                for (o,s,m) in ont.xref_graph.edges(data=True):
                    self.xref_graph.add_edge(o,s,**m)
            if ont.all_logical_definitions is not None:
                for ld in ont.all_logical_definitions:
                    self.add_logical_definition(ld)
            if ont.all_property_chain_axioms is not None:
                for pca in ont.all_property_chain_axioms:
                    self.add_property_chain_axiom(pca)

    def subgraph(self, nodes=None):
        """
        Return an induced subgraph

        By default this wraps networkx subgraph,
        but this may be overridden in specific implementations
        """
        if nodes is None:
            nodes = []
        return self.get_graph().subgraph(nodes)

    def subontology(self, nodes=None, minimal=False, relations=None):
        """
        Return a new ontology that is an extract of this one

        Arguments
        ---------
        - nodes: list
            list of node IDs to include in subontology. If None, all are used
        - relations: list
            list of relation IDs to include in subontology. If None, all are used

        """
        g = None
        if nodes is not None:
            g = self.subgraph(nodes)
        else:
            g = self.get_graph()
        if minimal:
            from ontobio.slimmer import get_minimal_subgraph
            g = get_minimal_subgraph(g, nodes)

        ont = Ontology(graph=g, xref_graph=self.xref_graph) # TODO - add metadata
        if relations is not None:
            g = ont.get_filtered_graph(relations)
            ont = Ontology(graph=g, xref_graph=self.xref_graph)
        return ont

    def create_slim_mapping(self, subset=None, subset_nodes=None, relations=None, disable_checks=False):
        """
        Create a dictionary that maps between all nodes in an ontology to a subset

        Arguments
        ---------
        ont : `Ontology`
            Complete ontology to be mapped. Assumed pre-filtered for relationship types
        subset : str
            Name of subset to map to, e.g. goslim_generic
        nodes : list
            If no named subset provided, subset is passed in as list of node ids
        relations : list
            List of relations to filter on
        disable_checks: bool
            Unless this is set, this will prevent a mapping being generated with non-standard relations.
            The motivation here is that the ontology graph may include relations that it is inappropriate to
            propagate gene products over, e.g. transports, has-part

        Return
        ------
        dict
            maps all nodes in ont to one or more non-redundant nodes in subset

        Raises
        ------
        ValueError
            if the subset is empty
        """
        if subset is not None:
            subset_nodes = self.extract_subset(subset)
            logger.info("Extracting subset: {} -> {}".format(subset, subset_nodes))

        if subset_nodes is None or len(subset_nodes) == 0:
            raise ValueError("subset nodes is blank")
        subset_nodes = set(subset_nodes)
        logger.debug("SUBSET: {}".format(subset_nodes))

        # Use a sub-ontology for mapping
        subont = self
        if relations is not None:
            subont = self.subontology(relations=relations)

        if not disable_checks:
            for r in subont.relations_used():
                if r != 'subClassOf' and r != 'BFO:0000050' and r != 'subPropertyOf':
                    raise ValueError("Not safe to propagate over a graph with edge type: {}".format(r))

        m = {}
        for n in subont.nodes():
            ancs = subont.ancestors(n, reflexive=True)
            ancs_in_subset = subset_nodes.intersection(ancs)
            m[n] = list(subont.filter_redundant(ancs_in_subset))
        return m

    def filter_redundant(self, ids):
        """
        Return all non-redundant ids from a list
        """
        sids = set(ids)
        for id in ids:
            sids = sids.difference(self.ancestors(id, reflexive=False))
        return sids

    def extract_subset(self, subset, contract=True):
        """
        Return all nodes in a subset.

        We assume the oboInOwl encoding of subsets, and subset IDs are IRIs, or IR fragments
        """
        return [n for n in self.nodes() if subset in self.subsets(n, contract=contract)]

    def subsets(self, nid, contract=True):
        """
        Retrieves subset ids for a class or ontology object
        """
        n = self.node(nid)
        subsets = []
        meta = self._meta(nid)
        if 'subsets' in meta:
            subsets = meta['subsets']
        else:
            subsets = []
        if contract:
            subsets = [self._contract_subset(s) for s in subsets]
        return subsets

    def _contract_subset(self, s):
        if s.find("#") > -1:
            return s.split('#')[-1]
        else:
            return s

    def _meta(self, nid):
        n = self.node(nid)
        return n.get("meta", {})

    def prefixes(self):
        """
        list all prefixes used
        """
        pset = set()
        for n in self.nodes():
            pfx = self.prefix(n)
            if pfx is not None:
                pset.add(pfx)
        return list(pset)

    def prefix_fragment(self, nid):
        """
        Return prefix and fragment/localid for a node
        """
        sep=':'
        if nid.startswith('http'):
            if '#' in nid:
               sep='#'
            else:
                sep='/'
        parts = nid.split(sep)
        frag = parts.pop()
        prefix = sep.join(parts)
        return prefix, frag

    def prefix(self, nid):
        """
        Return prefix for a node
        """
        pfx,_ = self.prefix_fragment(nid)
        return pfx

    def nodes(self):
        """
        Return all nodes in ontology

        Wraps networkx by default
        """
        return self.get_graph().nodes()

    def node(self, id):
        """
        Return a node with a given ID. If the node with the ID exists the
        Node object is returned, otherwise None is returned.

        Wraps networkx by default
        """
        return self.get_graph().nodes.get(id, None)

    def has_node(self, id):
        """
        True if id identifies a node in the ontology graph
        """
        return id in self.get_graph().nodes()

    def sorted_nodes(self):
        """
        Returns all nodes in ontology, after topological sort

        """
        return nx.topological_sort(self.get_graph())

    def node_type(self, id):
        """
        If stated, either CLASS, PROPERTY or INDIVIDUAL
        """
        return self.node(id).get('type', None)

    def relations_used(self):
        """
        Return list of all relations used to connect edges
        """
        g = self.get_graph()
        types = set()
        for (x,y,d) in g.edges(data=True):
            types.add(d['pred'])
        return list(types)

    def neighbors(self, node, relations=None):
        return self.parents(node, relations=relations) + self.children(node, relations=relations)

    def child_parent_relations(self, subj, obj, graph=None):
        """
        Get all relationship type ids between a subject and a parent.

        Typically only one relation ID returned, but in some cases there may be more than one

        Arguments
        ---------
        subj: string
            Child (subject) id
        obj: string
            Parent (object) id

        Returns
        -------
        list
        """
        if graph is None:
            graph = self.get_graph()
        preds = set()
        for _,ea in graph[obj][subj].items():
            preds.add(ea['pred'])
        logger.debug('{}->{} = {}'.format(subj,obj,preds))
        return preds

    def parents(self, node, relations=None):
        """
        Return all direct 'parents' of specified node.

        Note that in the context of ontobio, 'parent' means any node that
        is traversed in a single hop along an edge from a subject to object.
        For example, if the ontology has an edge "finger part-of some hand", then
        "hand" is the parent of finger.
        This can sometimes be counter-intutitive, for example, if the ontology
        contains has-part axioms. If the ontology has an edge
        "X receptor activity has-part some X binding", then "X binding" is the 'parent'
        of "X receptor activity" over a has-part edge.

        Wraps networkx by default.

        Arguments
        ---------
        node: string
           identifier for node in ontology
        relations: list of strings
           list of relation (object property) IDs used to filter

        """
        g = self.get_graph()
        if node in g:
            parents = list(g.predecessors(node))
            if relations is None:
                return parents
            else:
                rset = set(relations)
                return [p for p in parents if len(self.child_parent_relations(node, p, graph=g).intersection(rset)) > 0 ]
        else:
            return []

    def children(self, node, relations=None):
        """
        Return all direct children of specified node.

        Wraps networkx by default.

        Arguments
        ---------

        node: string

           identifier for node in ontology

        relations: list of strings

           list of relation (object property) IDs used to filter

        """
        g = self.get_graph()
        if node in g:
            children = list(g.successors(node))
            if relations is None:
                return children
            else:
                rset = set(relations)
                return [c for c in children if len(self.child_parent_relations(c, node, graph=g).intersection(rset)) > 0 ]
        else:
            return []

    def ancestors(self, node, relations=None, reflexive=False):
        """Return all ancestors of specified node.

        The default implementation is to use networkx, but some
        implementations of the Ontology class may use a database or
        service backed implementation, for large graphs.

        Arguments
        ---------
        node : str
            identifier for node in ontology
        reflexive : bool
            if true, return query node in graph
        relations : list
             relation (object property) IDs used to filter

        Returns
        -------
        list[str]
            ancestor node IDs

        """
        seen = set()
        nextnodes = [node]
        while len(nextnodes) > 0:
            nn = nextnodes.pop()
            if not nn in seen:
                seen.add(nn)
                nextnodes += self.parents(nn, relations=relations)
        if not reflexive:
            seen -= {node}
        return list(seen)

    def descendants(self, node, relations=None, reflexive=False):
        """
        Returns all descendants of specified node.

        The default implementation is to use networkx, but some
        implementations of the Ontology class may use a database or
        service backed implementation, for large graphs.


        Arguments
        ---------
        node : str
            identifier for node in ontology
        reflexive : bool
            if true, return query node in graph
        relations : list
             relation (object property) IDs used to filter

        Returns
        -------
        list[str]
            descendant node IDs
        """
        seen = set()
        nextnodes = [node]
        while len(nextnodes) > 0:
            nn = nextnodes.pop()
            if not nn in seen:
                seen.add(nn)
                nextnodes += self.children(nn, relations=relations)
        if not reflexive:
            seen -= {node}
        return list(seen)


    def equiv_graph(self):
        """
        Returns
        -------
        graph
            bidirectional networkx graph of all equivalency relations
        """
        eg = nx.Graph()
        for (u,v,d) in self.get_graph().edges(data=True):
            if d['pred'] == 'equivalentTo':
                eg.add_edge(u,v)
        return eg


    def traverse_nodes(self, qids, up=True, down=False, **args):
        """
        Traverse (optionally) up and (optionally) down from an input set of nodes

        Arguments
        ---------
        qids : list[str]
            list of seed node IDs to start from
        up : bool
            if True, include ancestors
        down : bool
            if True, include descendants
        relations : list[str]
            list of relations used to filter

        Return
        ------
        list[str]
            nodes reachable from qids
        """
        g = self.get_filtered_graph(**args)
        nodes = set()
        for id in qids:
            # reflexive - always add self
            nodes.add(id)
            if down:
                nodes.update(nx.descendants(g, id))
            if up:
                nodes.update(nx.ancestors(g, id))
        return nodes

    def get_roots(self, relations=None, prefix=None):
        """
        Get all nodes that lack parents

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter
        prefix : str
            E.g. GO. Exclude nodes that lack this prefix when testing parentage
        """
        g = self.get_filtered_graph(relations=relations, prefix=prefix)
        # note: we also eliminate any singletons, which includes obsolete classes
        roots = [n for n in g.nodes() if len(list(g.predecessors(n))) == 0 and len(list(g.successors(n))) > 0]
        return roots

    def get_level(self, level, relations=None, **args):
        """
        Get all nodes at a particular level

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter
        """
        g = self.get_filtered_graph(relations)
        nodes = self.get_roots(relations=relations, **args)
        for i in range(level):
            logger.info(" ITERATING TO LEVEL: {} NODES: {}".format(i, nodes))
            nodes = [c for n in nodes
                     for c in g.successors(n)]
        logger.info(" FINAL: {}".format(nodes))
        return nodes

    def parent_index(self, relations=None):
        """
        Returns a mapping of nodes to all direct parents

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter

        Returns:
        list
            list of lists [[CLASS_1, PARENT_1,1, ..., PARENT_1,N], [CLASS_2, PARENT_2,1, PARENT_2,2, ... ] ... ]
        """
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            g = self.get_filtered_graph(relations)
        l = []
        for n in g:
            l.append([n] + list(g.predecessors(n)))
        return l

    def text_definition(self, nid):
        """
        Retrieves logical definitions for a class or relation id

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried

        Returns
        -------
        TextDefinition
        """
        tdefs = []
        meta = self._meta(nid)
        if 'definition' in meta:
            obj = meta['definition']
            return TextDefinition(nid, **obj)
        else:
            return None


    def logical_definitions(self, nid):
        """
        Retrieves logical definitions for a class id

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried

        Returns
        -------
        LogicalDefinition
        """
        ldefs = self.all_logical_definitions
        if ldefs is not None:
            #print("TESTING: {} AGAINST LD: {}".format(nid, str(ldefs)))
            return [x for x in ldefs if x.class_id == nid]
        else:
            return []

    def add_logical_definition(self, logical_def):
        if self.all_logical_definitions is None:
            self.all_logical_definitions = []
        self.all_logical_definitions.append(logical_def)

    def get_property_chain_axioms(self, nid):
        """
        Retrieves property chain axioms for a class id

        Arguments
        ---------
        nid : str
             Node identifier for relation to be queried

        Returns
        -------
        PropertyChainAxiom
        """
        pcas = self.all_property_chain_axioms
        if pcas is not None:
            return [x for x in pcas if x.predicate_id == nid]
        else:
            return []

    def add_property_chain_axiom(self, pca):
        if self.all_property_chain_axioms is None:
            self.all_property_chain_axioms = []
        self.all_property_chain_axioms.append(pca)

    def get_node_type(self, nid):
        n = self.node(nid)
        if 'type' in n:
            return n['type']
        return None

    def _get_meta_prop(self, nid, prop):
        n = self.node(nid)
        if 'meta' in n:
            meta = n['meta']
            if prop in meta:
                return meta[prop]
        return None

    def _get_meta(self, nid):
        n = self.node(nid)
        if 'meta' in n:
            return n['meta']
        return None

    def _get_basic_property_values(self, nid):
        r =  self._get_meta_prop(nid, 'basicPropertyValues')
        if r is None:
            return []
        else:
            return r

    def _get_basic_property_value(self, nid, prop):
        bpvs = self._get_basic_property_values(nid)
        return [x['val'] for x in bpvs if x['pred'] == prop]

    def is_obsolete(self, nid):
        """
        True if node is obsolete

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        """
        dep = self._get_meta_prop(nid, 'deprecated')
        return  dep is not None and dep


    def replaced_by(self, nid, strict=True):
        """
        Returns value of 'replaced by' (IAO_0100001) property for obsolete nodes

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        strict: bool
            If true, raise error if cardinality>1. If false, return list if cardinality>1

        Return
        ------
        None if no value set, otherwise returns node id (or list if multiple values, see strict setting)
        """
        vs = self._get_basic_property_value(nid, 'IAO:0100001')
        if len(vs) > 1:
            msg = "replaced_by has multiple values: {}".format(vs)
            if strict:
                raise ValueError(msg)
            else:
                logger.error(msg)

        return vs

    def synonyms(self, nid, include_label=False):
        """
        Retrieves synonym objects for a class

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        include_label : bool
            If True, include label/names as Synonym objects

        Returns
        -------
        list[Synonym]
            :class:`Synonym` objects
        """
        n = self.node(nid)
        syns = []
        if 'meta' in n:
            meta = n['meta']
            if 'synonyms' in meta:
                for obj in meta['synonyms']:
                    syns.append(Synonym(nid, **obj))
        if include_label:
            syns.append(Synonym(nid, val=self.label(nid), pred='label'))
        return syns

    def obo_namespace(self, nid):
        go_namespace = [predval for predval in
                        self.get_graph().nodes
                            .get(nid, {})
                            .get("meta", {})
                            .get("basicPropertyValues", []) if predval["pred"] == "OIO:hasOBONamespace"]

        if len(go_namespace) >= 1:
            return go_namespace[0]["val"]
        else:
            return None


    def add_node(self, id, label=None, type='CLASS', meta=None):
        """
        Add a new node to the ontology
        """
        g = self.get_graph()
        if meta is None:
            meta={}
        g.add_node(id, label=label, type=type, meta=meta)

    def add_text_definition(self, textdef):
        """
        Add a new text definition to the ontology
        """
        self._add_meta_element(textdef.subject, 'definition', textdef.as_dict())

    def set_obsolete(self, nid):
        if nid not in self.get_graph():
            self.add_node(nid)
        self._add_meta_element(nid, 'deprecated', True)

    def _add_meta_element(self, id, k, edict):
        n = self.node(id)
        if n is None:
            raise ValueError('no such node {}'.format(id))
        if 'meta' not in n:
            n['meta'] = {}
        n['meta'][k] = edict

    def inline_xref_graph(self):
        """
        Copy contents of xref_graph to inlined meta object for each node
        """
        xg = self.xref_graph
        for n in self.nodes():
            if n in xg:
                self._add_meta_element(n, 'xrefs', [{'val':x} for x in xg.neighbors(n)])

    def add_parent(self, id, pid, relation='subClassOf'):
        """
        Add a new edge to the ontology
        """
        g = self.get_graph()
        g.add_edge(pid, id, pred=relation)

    def add_xref(self, id, xref):
        """
        Adds an xref to the xref graph
        """
        # note: does not update meta object
        if self.xref_graph is None:
            self.xref_graph = nx.MultiGraph()
        self.xref_graph.add_edge(xref, id)

    def add_synonym(self, syn):
        """
        Adds a synonym for a node
        """
        n = self.node(syn.class_id)
        if 'meta' not in n:
            n['meta'] = {}
        meta = n['meta']
        if 'synonyms' not in meta:
            meta['synonyms'] = []
        meta['synonyms'].append(syn.as_dict())

    def add_to_subset(self, id, s):
        """
        Adds a node to a subset
        """
        n = self.node(id)
        if 'meta' not in n:
            n['meta'] = {}
        meta = n['meta']
        if 'subsets' not in meta:
            meta['subsets'] = []
        meta['subsets'].append(s)

    def all_synonyms(self, include_label=False):
        """
        Retrieves all synonyms

        Arguments
        ---------
        include_label : bool
            If True, include label/names as Synonym objects

        Returns
        -------
        list[Synonym]
            :class:`Synonym` objects
        """
        syns = []
        for n in self.nodes():
            syns = syns + self.synonyms(n, include_label=include_label)
        return syns

    def all_obsoletes(self):
        """
        Returns all obsolete nodes
        """
        return [n for n in self.nodes() if self.is_obsolete(n)]

    def label(self, nid, id_if_null=False):
        """
        Fetches label for a node

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        id_if_null : bool
            If True and node has no label return id as label

        Return
        ------
        str
        """
        g = self.get_graph()
        if nid in g:
            n = g.nodes[nid]
            if 'label' in n:
                return n['label']
            else:
                if id_if_null:
                    return nid
                else:
                    return None
        else:
            if id_if_null:
                return nid
            else:
                return None

    def xrefs(self, nid, bidirectional=False, prefix=None):
        """
        Fetches xrefs for a node

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        bidirection : bool
            If True, include nodes xreffed to nid

        Return
        ------
        list[str]
        """
        xrefs = []
        if self.xref_graph is not None:
            xg = self.xref_graph
            if nid not in xg:
                xrefs = []
            else:
                if bidirectional:
                    xrefs = list(xg.neighbors(nid))
                else:
                    xrefs = [x for x in xg.neighbors(nid) if xg[nid][x][0]['source'] == nid]
                if prefix is not None:
                    xrefs = [x for x in xrefs if self.prefix(x) == prefix]
        return xrefs


    def resolve_names(self, names, synonyms=False, **args):
        """
        returns a list of identifiers based on an input list of labels and identifiers.

        Arguments
        ---------
        names: list
           search terms. '%' treated as wildcard
        synonyms: bool
           if true, search on synonyms in addition to labels
        is_regex : bool
           if true, treats each name as a regular expression
        is_partial_match : bool
           if true, treats each name as a regular expression .*name.*
        """
        g = self.get_graph()
        r_ids = []
        for n in names:
            logger.debug("Searching for {} syns={}".format(n,synonyms))
            if len(n.split(":")) == 2:
                r_ids.append(n)
            else:
                matches = set([nid for nid in g.nodes() if self._is_match(self.label(nid), n, **args)])
                if synonyms:
                    logger.debug("Searching syns for {}".format(names))
                    for nid in g.nodes():
                        for s in self.synonyms(nid):
                            if self._is_match(s.val, n, **args):
                                matches.add(nid)
                r_ids += list(matches)
        return r_ids

    def _is_match(self, label, term, is_partial_match=False, is_regex=False, **args):
        if term == '%':
            # always match if client passes '%'
            return True
        if label is None:
            label = ''
        if term.find('%') > -1:
            term = term.replace('%','.*')
            is_regex = True
        if is_regex:
            return re.search(term, label) is not None
        if is_partial_match:
            return label.find(term) > -1
        else:
            return label == term

    def search(self, searchterm, **args):
        """
        Simple search. Returns list of IDs.

        Arguments
        ---------
        searchterm: list
           search term. '%' treated as wildcard
        synonyms: bool
           if true, search on synonyms in addition to labels
        is_regex : bool
           if true, treats each name as a regular expression
        is_partial_match : bool
           if true, treats each name as a regular expression .*name.*

        Return
        ------
        list
           match node IDs
        """
        return self.resolve_names([searchterm], **args)

class LogicalDefinition():
    """
    A simple OWL logical definition conforming to the pattern:

    ::

        class_id = (genus_id_1 AND ... genus_id_n) AND (P_1 some FILLER_1) AND ... (P_m some FILLER_m)

    See `obographs <https://github.com/geneontology/obographs>`_ docs for more details

    """
    def __init__(self, class_id, genus_ids, restrictions):
        """
        Arguments
        ---------
        class_id : string
            the class that is being defined
        genus_ids : list
            a list of named classes (typically length 1)
        restrictions : list
            a list of (PROPERTY_ID, FILLER_CLASS_ID) tuples
        """
        self.class_id = class_id
        self.genus_ids = genus_ids
        self.restrictions = restrictions

    def __str__(self):
        return "{} =def {} AND {}".format(self.class_id, self.genus_ids, self.restrictions)
    def __repr__(self):
        return self.__str__()

class AbstractPropertyValue(object):
    """
    Abstract superclass of all property-value mapping classes.
    These correspond to Annotations in OWL
    """
    def __str__(self):
        return '{} "{}" {}'.format(self.subject, self.val, self.xrefs)
    def __repr__(self):
        return self.__str__()

    def __cmp__(self, other):
        (x,y) = (str(self),str(other))
        if x > y:
            return 1
        elif x < y:
            return -1
        else:
            return 0

class TextDefinition(AbstractPropertyValue):
    """
    Represents a textual definition for a class or relation
    """

    def __init__(self, subject, val=None, xrefs=None, ontology=None):
        """
        Arguments
        ---------
         - subject : string
             id for the class or relation that is being defined
         - val : string
             the definition itself
         - xrefs: list
             Provenance or cross-references to same usage
        """
        self.subject = subject
        self.val = val
        self.xrefs = xrefs
        self.ontology = ontology

    def as_dict(self):
        """
        Returns TextDefinition as obograph dict
        """
        return {
            'val': self.val,
            'xrefs': self.xrefs
        }

class Synonym(AbstractPropertyValue):
    """
    Represents a synonym using the OBO model
    """

    predmap = dict(
        label='label',
        hasExactSynonym='exact',
        hasBroadSynonym='broad',
        hasNarrowSynonym='narrow',
        hasRelatedSynonym='related')

    def __init__(self, class_id, val=None, pred='hasRelatedSynonym', lextype=None, xrefs=None, ontology=None, confidence=1.0, synonymType=None):
        """
        Arguments
        ---------
         - class_id : string
             the class that is being defined
         - val : string
             the synonym itself
         - pred: string
             oboInOwl predicate used to model scope. One of: has{Exact,Narrow,Related,Broad}Synonym - may also be 'label'
         - lextype: string
             From an open ended set of types
         - xrefs: list
             Provenance or cross-references to same usage

        """
        pred = pred.replace("http://www.geneontology.org/formats/oboInOwl#", "")
        self.class_id = class_id
        self.val = val
        self.pred = pred
        self.lextype = lextype
        self.xrefs = xrefs
        self.ontology = ontology
        self.confidence = confidence
        self.synonymType = synonymType

    def __str__(self):
        return '{} "{}" {} {} {} {} {}'.format(self.class_id, self.val, self.pred, self.lextype, self.xrefs, self.confidence, self.synonymType)

    def __repr__(self):
        return self.__str__()

    def scope(self):
        if self.pred in self.predmap:
            return self.predmap[self.pred].upper()
        else:
            return self.pred.upper()

    def is_label(self):
        return self.pred == 'label'

    def is_abbreviation(self, v=None):
        if v is not None:
            self.lextype = 'abbreviation'
        return self.lextype is not None and self.lextype.lower() == 'abbreviation'

    def exact_or_label(self):
        return self.pred == 'hasExactSynonym' or self.pred == 'label'

    def as_dict(self):
        """
        Returns Synonym as obograph dict
        """
        # TODO: complete metadata
        return {
            'pred': self.pred,
            'val': self.val,
            'xrefs': self.xrefs
        }

    def __cmp__(self, other):
        (x, y) = (str(self), str(other))
        if x > y:
            return 1
        if x < y:
            return -1
        return 0

class PropertyChainAxiom(object):
    """
    Represents a property chain axiom used to infer the existence of a property from a chain of properties.

    See the OWL primer for a description of property chains: https://www.w3.org/TR/owl2-primer/#Property_Chains
    """

    def __init__(self, predicate_id, chain_predicate_ids):
        """
        Arguments
        ---------
         - predicate_id : string
             the object property
         - chain_predicate_ids: list
             ordered list of chained properties

        """
        self.predicate_id = predicate_id
        self.chain_predicate_ids = chain_predicate_ids

    def __str__(self):
        return '{} ({})'.format(self.predicate_id, " o ".join(self.chain_predicate_ids))

    def as_dict(self):
        """
        Returns PropertyChainAxiom as obograph dict
        """
        return {
            "predicateId": self.predicate_id,
            "chainPredicateIds": self.chain_predicate_ids
        }



from ontobio.io import assocparser
from ontobio.io import parser_version_regex

from ontobio.model import association

from typing import Dict, List

import logging
import json

logger = logging.getLogger(__name__)


# TODO - use abstract parent for both entity and assoc
class EntityParser(assocparser.AssocParser):

    def parse(self, file, outfile=None):
        """Parse a line-oriented entity file into a list of entity dict objects

        Note the returned list is of dict objects. TODO: These will
        later be specified using marshmallow and it should be possible
        to generate objects

        Arguments
        ---------
        file : file or string
            The file is parsed into entity objects. Can be a http URL, filename or `file-like-object`, for input assoc file
        outfile : file
            Optional output file in which processed lines are written. This a file or `file-like-object`

        Return
        ------
        list
            Entities generated from the file
        """
        file = self._ensure_file(file)
        ents = []
        skipped = []
        n_lines = 0
        for line in file:
            n_lines += 1
            # if line.startswith("!"):
            #     if outfile is not None:
            #         outfile.write(line)
            #     continue
            line = line.strip("\n")
            if line == "":
                logger.warning("EMPTY LINE")
                continue

            parsed_line, new_ents = self.parse_line(line)
            if self._skipping_line(new_ents): # Skip if there were no ents
                logger.warning("SKIPPING: {}".format(line))
                skipped.append(line)
            else:
                for entity in new_ents:
                    if not entity.get("header", False):
                        # If this line is not a header (False if no "header" or if it's directly False) then add entity
                        ents += new_ents

                    # Always write out, so we'll copy the headers through
                    if outfile is not None:
                        outfile.write(parsed_line + "\n")

        self.report.skipped += len(skipped)
        self.report.n_lines += n_lines
        #self.report.n_associations += len(ents)
        logger.info("Parsed {} ents from {} lines. Skipped: {}".
                     format(len(ents),
                            n_lines,
                            len(skipped)))
        file.close()
        return ents

    def list_field(self, field: str) -> List:
        """
        Transforms a string field into a pipe separated list.
        Empty list if the input is the empty string.
        """
        return [] if field == "" else field.split("|")

# def load_gpi(self, gpi_path):
#     """
#     Loads a GPI as a file from the `config.gpi_authority_path`
#     """
#     if self.config.gpi_authority_path is not None:
#         gpis = dict()
#         parser = GpiParser()
#         with open(self.config.gpi_authority_path) as gpi_f:
#             entities = parser.parse(file=gpi_f)
#             for entity in entities:
#                 gpis[entity["id"]] = {
#                     "symbol": entity["label"],
#                     "name": entity["full_name"],
#                     "synonyms": entitywriter.stringify(entity["synonyms"]),
#                     "type": entity["type"]
#                 }
#         return gpis

    # If there is no config file path, return None
    # return None

class GpiParser(EntityParser):

    def __init__(self, config=None):
        """
        Arguments:
        ---------

        config : a assocparser.AssocParserConfig object
        """
        if config is None:
            config = assocparser.AssocParserConfig()
        self.config = config
        self.report = assocparser.Report()
        self.default_version = "1.2"
        self.version = None

    def gpi_version(self) -> str:
        if self.version:
            return self.version
        else:
            return self.default_version

    def parse_line(self, line):
        """Parses a single line of a GPI.

        Return a tuple `(processed_line, entities)`. Typically
        there will be a single entity, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """

        if self.is_header(line):
            if self.version is None:
                parsed = parser_version_regex.findall(line)
                if len(parsed) == 1:
                    filetype, version, _ = parsed[0]
                    if version == "2.0":
                        logger.info("Detected GPI version 2.0")
                        self.version = version
                    else:
                        logger.info("Detected version {}, so using 1.2".format(version))
                        self.version = self.default_version

            return (line, [{ "header": True, "line": line.strip() }])

        if self.version is None:
            logger.warning("No version number found for this file so we will assum GPI version: {}".format(self.default_version))
            self.version = self.default_version

        vals = line.split("\t")

        if len(vals) < 7:
            self.report.error(line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "")
            return line, []

        # If we are 1.2, then we can upconvert into a 2.0 "line", and validate from there
        if self.gpi_version() == "1.2":

            if len(vals) < 10 and len(vals) >= 7:
                missing_columns = 10 - len(vals)
                vals += ["" for i in range(missing_columns)]
            # Convert a 1.2 set of values to a 2.0 set of values
            vals = self.line_as_2_0(vals)
        else:
            # We are gpi 2.0
            if len(vals) < 11 and len(vals) >= 7:
                missing_columns = 11 - len(vals)
                vals += ["" for i in range(missing_columns)]

        vals = [el.strip() for el in vals]

        # End Housekeeping
        #=================================================================

        [
            object_id,
            db_object_symbol,
            db_object_name,
            synonyms,
            entity_types,
            taxon,
            encoded_by,
            parents,
            contained_complex_members,
            xrefs,
            properties
        ] = vals

        split_line = assocparser.SplitLine(line=line, values=vals, taxon=taxon)

        ## --
        ## db + db_object_id. CARD=1
        ## --
        if not self._validate_id(object_id, split_line):
            return line, []

        fullnames = self.list_field(db_object_name)

        ## --
        ## db_object_synonym CARD=0..*
        ## --
        synonyms = self.list_field(synonyms)

        types = self.list_field(entity_types)

        encoded_by = self.list_field(encoded_by)
        for encoded in encoded_by:
            self._validate_id(encoded, split_line)

        parents = [self._normalize_id(x) for x in self.list_field(parents)]
        for p in parents:
            self._validate_id(p, split_line)

        contained_complex_members = self.list_field(contained_complex_members)
        for members in contained_complex_members:
            self._validate_id(members, split_line)

        xref_ids = self.list_field(xrefs)

        obj = {
            'id': object_id,
            'label': db_object_symbol,
            'full_name': fullnames,
            'synonyms': synonyms,
            'type': types,
            'parents': parents,
            'encoded_by': encoded_by,
            'contained_complex_members': contained_complex_members,
            'xrefs': xref_ids,
            'taxon': {
                'id': self._taxon_id(taxon, split_line)
            },
            'properties': properties
        }
        return line, [obj]

    def line_as_2_0(self, columns) -> List[str]:
        """
        Convert a 1.2 line into a 2.0 line. Joins the db and object_id to form the
        final ID, uses empty string for unknown values in 2.0 spec (contained_complex_members
        and encoded_by) and otherwise forwards values onwards to the correct position in the
        2.0 line.
        """
        [
            db,
            db_object_id,
            db_object_symbol,
            db_object_name,
            db_object_synonym,
            db_object_type,
            taxon,
            parent_object_id,
            xrefs,
            properties
        ] = columns

        id = self._pair_to_id(db, db_object_id)

        as_2_0 = [
            id,
            db_object_symbol,
            db_object_name,
            db_object_synonym,
            db_object_type,
            taxon,
            "",
            parent_object_id,
            "",
            xrefs,
            properties
        ]
        return as_2_0

    def line_as_entity_subject(self, line: str):
        """
        This uses `parse_line` to produce the old-style dictionary for an Entity and
        then converts it to an `association.Subject` instance. This is essentially a
        shim for the newer `ontobio.model.collections` and `ontobio.model.association`
        modules to produce a `BioEntities` object from a GPI Parser.
        """
        # Parse the line first thing. This lets us parse header info and set the GPI
        # version state before we move on to the regular conversions.
        # If it's a header (starts with `!`) we'll just skip by returning None.
        _, entity_dicts = self.parse_line(line)

        if line.startswith("!"):
            return None

        subjects = []
        for entity in entity_dicts:
            entity_types = []
            if self.gpi_version() == "2.0":
                entity_types = [association.Curie.from_str(t) for t in entity["type"]]
                if any(c.is_error() for c in entity_types):
                    logger.error("Skipping `{}` due to malformed CURIE in entity type: `{}`".format(line, entity["type"]))
                    return None
            else:
                entity_types = entity["type"]

            parents = [association.Curie.from_str(p) for p in entity["parents"]]
            if any(p.is_error() for p in parents):
                logger.error("Skipping `{}` due to malformed CURIE in parents: `{}`".format(line, entity["parents"]))
                return None

            db_xrefs = [association.Curie.from_str(x) for x in entity["xrefs"]]
            if any(x.is_error() for x in db_xrefs):
                logger.error("Skipping `{}` due to malformed CURIE in db-xrefs: `{}`".format(line, entity["xrefs"]))
                return None

            complexes = [association.Curie.from_str(c) for c in entity["contained_complex_members"]]
            if any(c.is_error() for c in complexes):
                logger.error("Skipping `{}` due to malformed CURIE in contained complexes: `{}`".format(line, entity["contained_complex_members"]))
                return None

            s = association.Subject(
                id=association.Curie.from_str(entity["id"]), label=entity["label"], fullname=entity["full_name"],
                synonyms=entity["synonyms"], type=entity_types, taxon=association.Curie.from_str(entity["taxon"]["id"]),
                parents=parents, contained_complex_members=complexes, db_xrefs=db_xrefs, properties=entity["properties"])

            subjects.append(s)

        return subjects

class BgiParser(EntityParser):
    """
    BGI (basic gene info)
    """

    def __init__(self, config=None):
        """
        Arguments:
        ---------

        config : a assocparser.AssocParserConfig object
        """
        if config is None:
            config = assocparser.AssocParserConfig()
        self.config = config
        self.report = assocparser.Report()

    def parse(self, file, outfile=None):
        """Parse a BGI (basic gene info) JSON file
        """
        file = self._ensure_file(file)
        obj = json.load(file)
        items = obj['data']
        return [self.transform_item(item) for item in items]

    def transform_item(self, item):
        """
        Transforms JSON object
        """
        obj = {
            'id': item['primaryId'],
            'label': item['symbol'],
            'full_name': item['name'],
            'type': item['soTermId'],
            'taxon': {'id': item['taxonId']},
        }
        if 'synonyms' in item:
            obj['synonyms'] = item['synonyms']
        if 'crossReferenceIds' in item:
            obj['xrefs'] = [self._normalize_id(x) for x in item['crossReferenceIds']]

        # TODO: synonyms
        # TODO: genomeLocations
        # TODO: geneLiteratureUrl
        return obj



import json
import enum
import collections
import datetime
import copy
import logging

from dataclasses import dataclass

from typing import List, Dict, Any, Tuple, Union
from prefixcommons import curie_util
from ontobio import ontol
from ontobio import ecomap
from ontobio.io import assocparser
from ontobio.io import gaference
from ontobio.model import association
from ontobio.rdfgen import relations

logger = logging.getLogger(__name__)

FailMode = enum.Enum("FailMode", {"SOFT": "soft", "HARD": "hard"})
ResultType = enum.Enum("Result", {"PASS": "Pass", "WARNING": "Warning", "ERROR": "Error"})
RepairState = enum.Enum("RepairState", {"OKAY": "Okay", "REPAIRED": "Repaired", "FAILED": "Failed"})

# TestResult = collections.namedtuple("TestResult", ["result_type", "message", "result"])
class TestResult(object):
    def __init__(self, result_type: ResultType, message: str, result: List):
        self.result_type = result_type
        self.message = message
        self.result = result

"""
Send True for passes, and this returns the PASS ResultType, and if False, then
depending on the fail mode it returns either WARNING or ERROR ResultType.
"""
def result(passes: bool, fail_mode: FailMode) -> ResultType:
    if passes:
        return ResultType.PASS

    # Else we didn't pass
    if fail_mode == FailMode.SOFT:
        return ResultType.WARNING

    if fail_mode == FailMode.HARD:
        return ResultType.ERROR

def repair_result(repair_state: RepairState, fail_mode: FailMode) -> ResultType:
    if repair_state == RepairState.OKAY:
        return ResultType.PASS

    if repair_state == RepairState.REPAIRED:
        return ResultType.WARNING

    return result(False, fail_mode)


class GoRule(object):

    def __init__(self, id, title, fail_mode: FailMode, tags=[]):
        self.id = id
        self.title = title
        self.fail_mode = fail_mode
        self.run_context_tags = set(tags)

    def _list_terms(self, pipe_separated):
        terms = pipe_separated.split("|")
        terms = [t for t in terms if t != ""] # Remove empty strings
        return terms

    def _result(self, passes: bool) -> TestResult:
        return TestResult(result(passes, self.fail_mode), self.title, passes)

    def _is_run_from_context(self, config: assocparser.AssocParserConfig) -> bool:
        rule_tags_to_match = set([ "context-{}".format(c) for c in config.rule_contexts ])
        # If there is no context, then run
        # Or, if any run_context_tags is in rule_tags_to_match, then run
        return len(self.run_context_tags) == 0 or any(self.run_context_tags & rule_tags_to_match)

    def _run_if_context(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        result = TestResult(ResultType.PASS, "", annotation)
        if self._is_run_from_context(config):
            result = self.test(annotation, config, group=group)

        return result

    def run_test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        result = self._run_if_context(annotation, config, group=group)
        result.result = annotation
        return result


    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # run_test() -> _run_if_context() -> test()
        pass

class RepairRule(GoRule):

    def __init__(self, id, title, fail_mode, tags=[]):
        super().__init__(id, title, fail_mode, tags)

    def message(self, state: RepairState) -> str:
        message = ""
        if state == RepairState.REPAIRED:
            message = "Found violation of: `{}` but was repaired".format(self.title)
        elif state == RepairState.FAILED:
            message = "Found violatoin of: `{}` and could not be repaired".format(self.title)

        return message

    def run_test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        result = self._run_if_context(annotation, config, group=group)
        return result

    def repair(self, annotation: association.GoAssociation, group=None) -> Tuple[List, RepairState]:
        pass


class GoRule02(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000002", "No 'NOT' annotations to 'protein binding ; GO:0005515'", FailMode.SOFT)


    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:

        fails = (str(annotation.object.id) == "GO:0005515" and annotation.negated)
        return self._result(not fails)

class GoRule06(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000006", "IEP and HEP usage is restricted to terms from the Biological Process ontology", FailMode.HARD)
        self.iep = "ECO:0000270"
        self.hep = "ECO:0007007"

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        if config.ontology is None:
            return self._result(True)

        go_namespace = [predval for predval in config.ontology.get_graph().nodes.get(str(annotation.object.id), {}).get("meta", {}).get("basicPropertyValues", []) if predval["pred"]=="OIO:hasOBONamespace"]
        evidence = str(annotation.evidence.type)
        fails = evidence in [self.iep, self.hep] and "biological_process" not in [o["val"] for o in go_namespace]
        return self._result(not fails)

class GoRule07(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000007", "IPI should not be used with catalytic activity molecular function terms", FailMode.SOFT)
        self.children_of_catalytic_activity = None

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        catalytic_activity = "GO:0003824"
        if config.ontology is not None and self.children_of_catalytic_activity is None:
            # We'll define children_of_catalytic_activity if we have an ontology *and* if we haven't defined it before already
            self.children_of_catalytic_activity = set(config.ontology.descendants(catalytic_activity, relations=["subClassOf"], reflexive=True))

        goterm = str(annotation.object.id)
        evidence = str(annotation.evidence.type)

        fails = False
        if self.children_of_catalytic_activity is not None:
            # We fail if evidence is IPI and the goterm is a subclass of catalytic activity, else we good
            fails = evidence == "ECO:0000353" and goterm in self.children_of_catalytic_activity

        return self._result(not fails)


class GoRule08(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000008", "No annotations should be made to uninformatively high level terms", FailMode.SOFT)
        self.do_not_annotate = None
        self.do_not_manually_annotate = None

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # Cache the subsets
        if config.ontology is None:
            return self._result(True)

        if self.do_not_annotate is None and config.ontology is not None:
            self.do_not_annotate = set(config.ontology.extract_subset("gocheck_do_not_annotate"))
            self.do_not_manually_annotate = set(config.ontology.extract_subset("gocheck_do_not_manually_annotate"))
        elif self.do_not_annotate is None and config.ontology is None:
            self.do_not_annotate = []
            self.do_not_manually_annotate = []

        goid = str(annotation.object.id)
        evidence = str(annotation.evidence.type)

        auto_annotated = goid in self.do_not_annotate
        manually_annotated = evidence != "ECO:0000501" and goid in self.do_not_manually_annotate
        not_high_level = not (auto_annotated or manually_annotated)

        t = result(not_high_level, self.fail_mode)
        return TestResult(t, self.title, not_high_level)


class GoRule11(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000011", "ND annotations to root nodes only", FailMode.HARD)
        self.root_go_classes = ["GO:0003674", "GO:0005575", "GO:0008150"]
        self.nd = "ECO:0000307"

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        goclass = str(annotation.object.id)
        evidence = str(annotation.evidence.type)

        # If we see a bad evidence, and we're not in a paint file then fail.
        # We're good if both predicates are true, or neither are true
        success = (evidence == self.nd and goclass in self.root_go_classes) or (evidence != self.nd and goclass not in self.root_go_classes)
        return self._result(success)

class GoRule13(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000013", "Taxon-appropriate annotation check", FailMode.HARD)
        self.non_experimental_evidence = set(["ECO:0000318", "ECO:0000320", "ECO:0000321", "ECO:0000305", "ECO:0000247", "ECO:0000255", "ECO:0000266", "ECO:0000250", "ECO:0000303", "ECO:0000245", "ECO:0000304", "ECO:0000307", "ECO:0000501"])

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        if config.annotation_inferences is None:
            # Auto pass if we don't have inferences
            return self._result(True)

        if annotation.negated:
            # The rule is passed if the annotation is negated
            return self._result(True)

        inference_results = gaference.produce_inferences(annotation, config.annotation_inferences)  # type: List[gaference.InferenceResult]
        taxon_passing = True
        for result in inference_results:
            if result.problem == gaference.ProblemType.TAXON:
                taxon_passing = False
                break

        if taxon_passing:
            return self._result(True)
        else:
            # Filter non experimental evidence
            if str(annotation.evidence.type) in self.non_experimental_evidence:
                return self._result(False)
            else:
                # Only submit a warning/report if we are an experimental evidence
                return TestResult(ResultType.WARNING, self.title, False)

class GoRule15(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000015", "Dual species taxon check", FailMode.SOFT)
        self.allowed_dual_species_terms = None

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:

        # Cache the allowed terms
        if self.allowed_dual_species_terms is None and config.ontology is not None:
            interaction_terms = config.ontology.descendants("GO:0044419", relations=["subClassOf"], reflexive=True)
            interspecies_interactions_regulation = config.ontology.descendants("GO:0043903", relations=["subClassOf"], reflexive=True)
            host_cellular_component = config.ontology.descendants("GO:0018995", relations=["subClassOf"], reflexive=True)
            self.allowed_dual_species_terms = set(interaction_terms + interspecies_interactions_regulation + host_cellular_component)
        elif config.ontology is None:
            return self._result(True)

        passes = False
        if self.allowed_dual_species_terms is not None:
            dual = annotation.interacting_taxon is not None
            goterm = str(annotation.object.id)

            # We fail if we are a dual taxon and then the term is not in this list
            # This is the same as dual -> goterm in list
            # Implication rewritten is Not P OR Q
            passes = not dual or (goterm in self.allowed_dual_species_terms)

        return self._result(passes)


class GoRule16(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000016", "All IC annotations should include a GO ID in the \"With/From\" column", FailMode.HARD)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)
        withfrom = annotation.evidence.with_support_from

        okay = True
        if evidence == "ECO:0000305":
            only_go = [t for conjunctions in withfrom for t in conjunctions.elements if t.namespace == "GO"] # Filter terms that aren't GO terms
            okay = len(only_go) >= 1

        return self._result(okay)


class GoRule17(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000017", "IDA annotations must not have a With/From entry", FailMode.SOFT)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)
        withfrom = annotation.evidence.with_support_from

        if evidence == "ECO:0000314":
            return self._result(not bool(withfrom))
        else:
            return self._result(True)

class GoRule18(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000018", "IPI annotations require a With/From entry", FailMode.SOFT)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)
        withfrom = annotation.evidence.with_support_from

        if evidence == "ECO:0000353":
            return self._result(bool(withfrom))
        else:
            return self._result(True)


class GoRule26(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000026", "IBA evidence codes should be filtered from main MOD gaf sources", FailMode.HARD)
        self.offending_evidence = ["ECO:0000318"]

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)
        # If we see a bad evidence, and we're not in a paint file then fail.
        fails = (evidence in self.offending_evidence and not config.paint)
        return self._result(not fails)

class GoRule28(RepairRule):
    def __init__(self):
        super().__init__("GORULE:0000028", "Aspect can only be one of C, P, F", FailMode.HARD)
        self.namespace_aspect_map = {
            "biological_process": "P",
            "cellular_component": "C",
            "molecular_function": "F"
        }

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        aspect = annotation.aspect
        goterm = str(annotation.object.id)

        if config.ontology is None:
            return TestResult(ResultType.PASS, self.title, annotation)

        namespaces = [predval for predval in config.ontology.get_graph().nodes.get(goterm, {}).get("meta", {}).get("basicPropertyValues", []) if predval["pred"]=="OIO:hasOBONamespace"]
        # the namespaces expression cascades through the json representation of this
        # ontology using empty dict/list if the key is not present

        if len(namespaces) == 0:
            # If this doesn't exist, then it's fine
            return TestResult(ResultType.PASS, self.title, annotation)

        namespace = namespaces[0]["val"]
        expected_aspect = self.namespace_aspect_map[namespace]

        correct_aspect = expected_aspect == aspect
        annotation.aspect = expected_aspect

        repair_state = None
        if correct_aspect:
            repair_state = RepairState.OKAY
        else:
            repair_state = RepairState.REPAIRED

        return TestResult(repair_result(repair_state, self.fail_mode), self.message(repair_state), annotation)


class GoRule29(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000029", "All IEAs over a year are warned, all IEAs over two are removed", FailMode.HARD)
        self.one_year = datetime.timedelta(days=365)
        self.two_years = datetime.timedelta(days=730)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)
        date = annotation.date

        now = datetime.datetime.today()

        time_compare_delta_short = self.one_year
        time_compare_delta_long = self.two_years
        time_diff = now - datetime.datetime(int(date.year),
                                            int(date.month),
                                            int(date.day),
                                            0, 0, 0, 0)

        iea = "ECO:0000501"
        if evidence == iea:
            if time_diff > time_compare_delta_long:
                return self._result(False)
            elif time_diff > time_compare_delta_short:
                return TestResult(ResultType.WARNING, self.title, annotation)

        ## Default results we we get here.
        return self._result(True)

class GoRule30(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000030", "Deprecated GO_REFs are not allowed", FailMode.SOFT)

    def _ref_curi_to_id(self, goref) -> str:
        """
        Changes reference IDs in the form of GO_REF:nnnnnnn to goref-nnnnnnn
        """
        return goref.lower().replace("_", "").replace(":", "-")

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        references = annotation.evidence.has_supporting_reference
        for ref in references:
            ref = str(ref)
            # Not allowed is obsolete and GO_PAINT:x
            if ref.startswith("GO_PAINT") or (config.goref_metadata is not None and config.goref_metadata.get(self._ref_curi_to_id(ref), {}).get("is_obsolete", False)):
                return self._result(False)

        return self._result(True)

class GoRule37(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000037", "IBA annotations should ONLY be assigned_by GO_Central and have PMID:21873635 as a reference", FailMode.HARD)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # If the evidence code is IBA, then (1) the assigned_by field must be GO_Central and (2) the reference field must be PMID:21873635
        evidence = str(annotation.evidence.type)
        references = [str(ref) for ref in annotation.evidence.has_supporting_reference]
        assigned_by = annotation.provided_by

        result = self._result(True) # By default we pass
        if evidence == "ECO:0000318":
            result = self._result(assigned_by == "GO_Central" and "PMID:21873635" in references)

        return result

class GoRule39(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000039", "Protein complexes can not be annotated to GO:0032991 (protein-containing complex) or its descendants", FailMode.HARD)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # An implementation note: This is done by testing if the DB (column 1) is ComplexPortal.
        # This will grab a subset of all actual Protein Complexes. This is noted in the rule description
        db = annotation.subject.id.namespace
        goterm = str(annotation.object.id)

        fails = (db == "ComplexPortal" and goterm == "GO:0032991")
        return self._result(not fails)

class GoRule42(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000042", "Qualifier: IKR evidence code requires a NOT qualifier", FailMode.HARD)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        evidence = str(annotation.evidence.type)

        result = self._result(True)
        ikr = "ECO:0000320"
        if evidence == ikr:
            result = self._result(annotation.negated)

        return result

class GoRule43(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000043", "Check for valid combination of evidence code and GO_REF", FailMode.SOFT)
        self.ecomapping = ecomap.EcoMap()

    def _ref_curi_to_id(self, goref) -> str:
        """
        Changes reference IDs in the form of GO_REF:nnnnnnn to goref-nnnnnnn
        """
        return goref.lower().replace("_", "").replace(":", "-")

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        if config.goref_metadata is None:
            return self._result(True)

        references = [str(ref) for ref in annotation.evidence.has_supporting_reference]
        evidence = str(annotation.evidence.type)

        for ref in references:
            allowed_eco = config.goref_metadata.get(self._ref_curi_to_id(ref), {}).get("evidence_codes", None)
            # allowed_eco are ECO curies
            # allowed_eco will only not be none if the ref was GO_REF:nnnnnnn, that's the only time we care here
            if allowed_eco:
                if evidence not in allowed_eco:
                    return self._result(False)

        return self._result(True)


class GoRule46(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000046", "The ‘with’ field (GAF column 8) must be the same as the gene product (GAF colummn 2) when annotating to ‘self-binding’ terms", FailMode.SOFT)
        self.self_binding_roots = ["GO:0042803", "GO:0051260", "GO:0051289", "GO:0070207", "GO:0043621", "GO:0032840"]
        self.self_binding_terms = None

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        if config.ontology is not None and self.self_binding_terms is None:
            all_terms = []
            # Initialize the self_binding terms if we have an ontology and we haven't already initialized the terms
            for binding_root in self.self_binding_roots:
                root_descendants = config.ontology.descendants(binding_root, relations=["subClassOf"], reflexive=True)
                all_terms += root_descendants

            self.self_binding_terms = set(all_terms)
        elif config.ontology is None:
            # Make sure if we don't have an ontology we still use the set roots
            self.self_binding_terms = self.self_binding_roots

        withfroms = annotation.evidence.with_support_from
        goterm = str(annotation.object.id)

        if goterm in self.self_binding_terms:
            # Then we're in the self-binding case, and check if object ID is in withfrom
            for conj in withfroms:
                if annotation.subject.id in conj.elements:
                    return self._result(True)

            return self._result(False)

        return self._result(True)

class GoRule50(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000050", "Annotations to ISS, ISA and ISO should not be self-referential", FailMode.SOFT)
        self.the_evidences = ["ECO:0000250", "ECO:0000247", "ECO:0000266"]

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # should not have the same identifier in the 'gene product column' (column 2) and in the 'with/from' column (column 8)
        evidence = str(annotation.evidence.type)
        result = self._result(True)
        if evidence in self.the_evidences:
            # Ensure the gp ID is not an entry in withfrom
            for conj in annotation.evidence.with_support_from:
                result = self._result(annotation.subject.id not in conj.elements)

        return result

class GoRule55(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000055", "References should have only one ID per ID space", FailMode.SOFT)

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        found_id_spaces = dict()
        for ref in annotation.evidence.has_supporting_reference:
            if ref.namespace in found_id_spaces:
                return self._result(False)
            else:
                found_id_spaces[ref.namespace] = ref
        # We found no duplicate IDs, so we good
        return self._result(True)


class GoRule57(GoRule):

    def __init__(self):
        super().__init__("GORULE:0000057", "Group specific filter rules should be applied to annotations", FailMode.HARD, tags=["context-import"])

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:
        # Check group_metadata is present
        if config.group_metadata is None:
            return self._result(True)

        evidence_codes = config.group_metadata.get("filter_out", {}).get("evidence", [])
        if str(annotation.evidence.type) in evidence_codes:
            return self._result(False)

        evidences_references = config.group_metadata.get("filter_out", {}).get("evidence_reference", [])
        for er in evidences_references:
            evidence_code = er["evidence"]
            reference = er["reference"]
            if str(annotation.evidence.type) == evidence_code and [str(ref) for ref in annotation.evidence.has_supporting_reference] == [reference]:
                return self._result(False)

        properties = config.group_metadata.get("filter_out", {}).get("annotation_properties", [])
        for p in properties:
            if p in annotation.properties.keys():
                return self._result(False)

        return self._result(True)

class GoRule58(RepairRule):

    def __init__(self):
        super().__init__("GORULE:0000058", "Object extensions should conform to the extensions-patterns.yaml file in metadata", FailMode.HARD, tags=["context-import"])

    def test(self, annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> TestResult:

        if config.extensions_constraints is None:
            return TestResult(ResultType.PASS, self.title, annotation)

        if config.ontology is None:
            return TestResult(ResultType.PASS, self.title, annotation)

        repair_state = RepairState.OKAY

        bad_conjunctions = []
        for con in annotation.object_extensions:
            # Count each extension unit, represented by tuple (Relation, Namespace)
            extension_counts = collections.Counter([(str(unit.relation), unit.term.namespace) for unit in con.elements])

            matches = self._do_conjunctions_match_constraint(con, annotation.object.id, config.extensions_constraints, extension_counts)
            # If there is a match in the constraints, then we're all good and we can exit with a pass!
            if not matches:
                bad_conjunctions.append(con)
                repair_state = RepairState.REPAIRED

        repaired_annotation = copy.deepcopy(annotation)
        for con in bad_conjunctions:
            # Remove the bad conjunctions as the "repair"
            repaired_annotation.object_extensions.remove(con)

        return TestResult(repair_result(repair_state, self.fail_mode), self.message(repair_state), repaired_annotation)

    """
    This matches a conjunction against the extension constraints passed in through `extensions-constraints.yaml` in go-site.
    The extensions constraints acts as a white list, and as such each extension unit in the conjunction must
    find a match in the constraints list: the extension relation must match a constraint, and if it does, the
    namespace of the extension filler must much an allowed namespace in the constraint, the annotation GO term
    must match one of the classes in 'primary_terms', and possibly a cardinality of (Relation, Namespace) must
    not be violated.

    If such a match is found, then we can move to the next extension unit in the conjunction list. If each extension has
    a match in the constraints then the conjunction passes the test.

    Any extension unit that fails means the entire conjunction fails.
    """
    def _do_conjunctions_match_constraint(self, conjunction, term, constraints, conjunction_counts):
        # Check each extension in the conjunctions
        for ext in conjunction.elements:

            extension_good = False
            for constraint in constraints:
                constraint_relation_uri = association.relations.lookup_label(constraint["relation"])
                ext_relation_uri = curie_util.expand_uri(str(ext.relation))
                if ext_relation_uri == constraint_relation_uri:

                    if (ext.term.namespace in constraint["namespaces"] and str(term) in constraint["primary_terms"]):
                        # If we match namespace and go term, then if we there is a cardinality constraint, check that.
                        if "cardinality" in constraint:
                            cardinality_violations = [(ext, num) for ext, num in dict(conjunction_counts).items() if num > constraint["cardinality"]]
                            extension_good = len(cardinality_violations) == 0
                        else:
                            extension_good = True

                        if extension_good:
                            # If things are good for this extension, break and go to the next one
                            break

            # if we get through all constraints and we found no constraint match for `ext`
            # Then we know that `ext` is wrong, making the whole conjunction wrong, and we can bail here.
            if not extension_good:
                return False

        # If we get to the end of all extensions without failing early, then the conjunction is good!
        return True


GoRules = enum.Enum("GoRules", {
    "GoRule02": GoRule02(),
    "GoRule06": GoRule06(),
    "GoRule07": GoRule07(),
    "GoRule08": GoRule08(),
    "GoRule11": GoRule11(),
    "GoRule15": GoRule15(),
    "GoRule16": GoRule16(),
    "GoRule17": GoRule17(),
    "GoRule18": GoRule18(),
    "GoRule26": GoRule26(),
    "GoRule28": GoRule28(),
    "GoRule29": GoRule29(),
    "GoRule30": GoRule30(),
    "GoRule37": GoRule37(),
    "GoRule39": GoRule39(),
    "GoRule42": GoRule42(),
    "GoRule43": GoRule43(),
    "GoRule46": GoRule46(),
    "GoRule50": GoRule50(),
    "GoRule55": GoRule55(),
    "GoRule57": GoRule57(),
    "GoRule58": GoRule58(),
    # GoRule13 at the bottom in order to make all other rules clean up an annotation before reaching 13
    "GoRule13": GoRule13()
})


GoRulesResults = collections.namedtuple("GoRulesResults", ["all_results", "annotation"])
def test_go_rules(annotation: association.GoAssociation, config: assocparser.AssocParserConfig, group=None) -> GoRulesResults:
    all_results = {}

    active_annotation = annotation
    for rule in list(GoRules):
        result = rule.value.run_test(active_annotation, config, group=group)
        # Accumulate all repairs performed  by all tests to the annotation
        active_annotation = result.result
        all_results[rule.value] = result

    return GoRulesResults(all_results, active_annotation)



from enum import Enum


class HomologyTypes(Enum):
    """
    Core homology relations from RO
    """
    Ortholog = 'RO:HOM0000017'
    LeastDivergedOrtholog = 'RO:HOM0000020'
    Homolog = 'RO:HOM0000007'
    Paralog = 'RO:HOM0000011'
    InParalog = 'RO:HOM0000023'
    OutParalog = 'RO:HOM0000024'
    Ohnolog = 'RO:HOM0000022'
    Xenolog = 'RO:HOM0000018'

class Evidence():
    axiom_has_evidence = 'RO:0002612'
    evidence_with_support_from =  'RO:0002614'
    has_supporting_reference =  'SEPIO:0000124'
    information_artefact = 'IAO:0000311'

    _prefixmap = {
        'SEPIO': 'http://purl.obolibrary.org/obo/SEPIO_',
        'IAO': 'http://purl.obolibrary.org/obo/IAO_',
        'RO': 'http://purl.obolibrary.org/obo/RO_',
    }
    
class OboRO():
    part_of = 'BFO:0000050'
    occurs_in = 'BFO:0000066'
    enabled_by = 'RO:0002333'
    enables = 'RO:0002327'
    involved_in = 'RO:0002331'
    in_taxon = 'RO:0002162'
    colocalizes_with = 'RO:0002325'

def map_legacy_pred(pred):
    if '#' in pred:
        lbl = pred.split('#')[-1]
        if lbl == 'part_of':
            return 'BFO:0000050'
        if '_part_of' in lbl:
            # for FMA
            return 'BFO:0000050'
    return pred
    



from ontobio.rdfgen.gocamgen.utils import contract_uri_wrapper

### Simple example: 'GP --enabled_by--> MF --part_of--> BP'
### First find all 'GP --enabled_by--> MF' triples
### Then find a 'MF --part_of--> BP' where URI of MF equals URI of MF of any of triples in first set.
### So subsequent links (triples) in the chain will be passed the whole set of candidate preceding chains
### As long as current triple matches next link in pattern and URI of subject matches URI of previous triple's object, keep going


class TriplePattern:
    def __init__(self, ordered_triples):
        self.ordered_triples = ordered_triples


class ConnectedTriplePattern(TriplePattern):
    def __init__(self, ordered_triples):
        TriplePattern.__init__(self, ordered_triples)


class TriplePair:
    def __init__(self, triple_a, triple_b, connecting_entity):
        self.triples = (triple_a, triple_b)
        # Check to make sure connecting_entity is in both triples
        self.connecting_entity = connecting_entity

    def is_connected_by_uri(self, model):
        #TODO Check that subjects and objects are URIs
        for uri in [self.triples[0][0], self.triples[0][2]]:
            class_curie = model.class_for_uri(uri)
            if class_curie == self.connecting_entity and uri in self.triples[1]:
                return True
        return False

    def connecting_entity_uri(self, model):
        # entity is either [0] or [2] of either self.triples[0] or self.triples[1]
        # Scan these, checking class_for_uri == connecting_entity
        for uri in [self.triples[0][0], self.triples[0][2], self.triples[1][0], self.triples[1][2]]:
            class_curie = model.class_for_uri(uri)
            if class_curie == self.connecting_entity:
                return uri
        return None

# AKA pattern? Holds all pairs and will be query input for TriplePatternFinder
class TriplePairCollection:
    def __init__(self):
        self.chain_collection = []


class TriplePatternFinder:

    def find_pattern_recursive(self, model, pattern: TriplePattern, candidate_chains=[], exact_length=False,
                               pattern_length=None):
        # break down pattern into component triples
        p = pattern.ordered_triples[0]
        found_triples = model.triples_by_ids(*p)

        if len(found_triples) > 0:
            if len(candidate_chains) == 0:
                # This is start of chain
                # candidate_chains = [[t] for t in found_triples]
                candidate_chains = []
                for t in found_triples:
                    if exact_length and len(pattern.ordered_triples) == 1:
                        if self.triple_individuals_only_in_chain(model, [t], t):
                            # stuff is gonna start on fire
                            fire = "started"
                            candidate_chains.append([t])
                    else:
                        candidate_chains.append([t])
                if pattern_length is None:
                    pattern_length = len(pattern.ordered_triples)
            else:
                candidate_chains_local = []
                for chain in candidate_chains:
                    for triple in found_triples:
                        #TODO: Make more selective (e.g. subject or object must be in previous triple)
                        if exact_length and len(pattern.ordered_triples) == 1:
                            # This is the last triple in chain. More checks needed. Triple connected to other triples?
                            subject_triples = model.triples_involving_individual(triple[0])
                            object_triples = model.triples_involving_individual(triple[2])
                            # Er, ANY other triples with subject or object?
                            # Any of these contain triples not in chain?
                            drop_chain = False
                            for t in subject_triples + object_triples:
                                if t not in chain:
                                    drop_chain = True
                            if not drop_chain:
                                candidate_chains_local.append(chain + [triple])
                        else:
                            candidate_chains_local.append(chain + [triple])
                candidate_chains = candidate_chains_local

            if len(pattern.ordered_triples) > 1:
                rest_of_pattern = TriplePattern(pattern.ordered_triples[1:len(pattern.ordered_triples)])
                candidate_chains = self.find_pattern_recursive(model, rest_of_pattern, candidate_chains,
                                                               exact_length=exact_length, pattern_length=pattern_length)
            elif exact_length:
                # On last triple of pattern. Throw out chains with len() != len(starting_pattern)
                candidate_chains_local = []
                for chain in candidate_chains:
                    # And end triple of chain
                    if len(chain) == pattern_length:
                        candidate_chains_local.append(chain)
                candidate_chains = candidate_chains_local
        else:
            candidate_chains = []

        return candidate_chains

    # def find_connected_pattern(self, model: AssocGoCamModel, pair_collection: TriplePairCollection, candidate_chains=[]):
    def find_connected_pattern(self, model, pair_collection: TriplePairCollection, exact=False):
        # pair_collection is a TriplePairCollection w/ chain_collection of TriplePair[]. Each of these TriplePairs
        # consists of two triples, e.g. ("MF-term", relation_uri, "GP-class")
        # Returned output should be a TriplePairCollection w/ chain_collection of TriplePairs consisting of all-URI triples.
        connected_pair_collection = TriplePairCollection()  # will be triples of URIs
        for pair in pair_collection.chain_collection:
            pattern = TriplePattern([*pair.triples])
            found_pairs = self.find_pattern_recursive(model, pattern)  # only using this 2-deep
            connected_pair = None
            for fpair in found_pairs:
                # fpair is list of two triples
                #TODO put into TriplePairs
                f_triple_pair = TriplePair(*fpair, connecting_entity=pair.connecting_entity)
                if f_triple_pair.is_connected_by_uri(model):
                    connected_pair = f_triple_pair
                    break
            if connected_pair:
                connected_pair_collection.chain_collection.append(connected_pair)
            else:
                # TODO: Should this return 'connected_pair_collection' instead?
                return None
        return connected_pair_collection

    def triple_individuals_only_in_chain(self, model, chain, triple):
        # This is the last triple in chain. More checks needed. Triple connected to other triples?
        subject_triples = model.triples_involving_individual(triple[0])
        object_triples = model.triples_involving_individual(triple[2])
        # Er, ANY other triples with subject or object?
        # Any of these contain triples not in chain?
        for t in subject_triples + object_triples:
            relation_curie = contract_uri_wrapper(t[1])[0]
            # We basically just want to look at RO, BFO relations
            if not (relation_curie.startswith("RO:") or relation_curie.startswith("BFO:")):
                continue
            if t not in chain:
                return False
        return True



from networkx import MultiDiGraph
from rdflib.term import URIRef
from ontobio.rdfgen.gocamgen.rdflib_sparql_wrapper import RdflibSparqlWrapper
from ontobio.rdfgen.gocamgen.utils import expand_uri_wrapper
from ontobio.rdfgen.gocamgen.errors import ModelRdfWriteException
import logging

logger = logging.getLogger(__name__)

# How do we track/compare fully built annotations before adding to model? We need a separate
# subgraph for each assertion (no IRIs, just classes)
# These subgraphs get parsed into SPARQL queries.
#
# Number each instance like so:
# GO:0003674-1 enabled_by WB:WBGene00001173-1
# GO:0003674-1 part_of GO:0051343-1
# GO:0003674-1 has_input WB:WBGene00001173-2
# GO:0003674-1 causally_upstream_of GO:0003674-2
# GO:0003674-2 enabled_by WB:WBGene00001173-2
# Are there any "-"s in identifiers?

class AnnotationSubgraph(MultiDiGraph):

    def __init__(self, annot):
        MultiDiGraph.__init__(self)
        # self.source_line = annot.get("source_line").rstrip().replace("\t", " ")
        self.class_counts = {}

    def add_edge(self, u_for_edge, relation, v_for_edge, key=None, **attr):
        attr["relation"] = relation
        MultiDiGraph.add_edge(self, u_for_edge, v_for_edge, key, **attr)

    # Will turn input of GO:0045944 into node_id=GO:0045944-1, sparql_var=GO_0045944_1
    # Will turn input of MGI:MGI:98956 into node_id=MGI:MGI:98956-1, sparql_var=MGI_MGI_98956_1
    # Ex:
    # >>> g.nodes['GO:0045944-1']
    # {'sparql_var': 'GO_0045944_1'}
    # Returns the networkx graph node ID used in tracking instances
    # Use this node ID to get instance IRI (once it's defined) and SPARQL variable name
    def add_instance_of_class(self, class_id, is_anchor=False):
        next_num = self.increment_class_count(class_id)
        node_id = "{}-{}".format(class_id, next_num)
        # TODO: Maybe sparql_var doesn't need to be stored and can always be parsed out of node_id? Make function
        sparql_var = node_id.replace(":", "_")
        sparql_var = sparql_var.replace("-", "_")
        self.add_node(node_id, sparql_var=sparql_var)
        if is_anchor:
            self.set_anchor(node_id)

        return node_id

    def increment_class_count(self, class_id):
        if class_id not in self.class_counts:
            self.class_counts[class_id] = 0
        self.class_counts[class_id] += 1
        return self.class_counts[class_id]

    # Anchor is used for tracking which instance to hang extensions from
    def set_anchor(self, node_id):
        # Reset all nodes
        for n in self:
            is_anchor = self.nodes[n].get("is_anchor")
            if is_anchor:
                self.nodes[n]["is_anchor"] = None
        # Mark this node_id as anchor
        self.nodes[node_id]["is_anchor"] = True

    def get_anchor(self):
        for n in self:
            is_anchor = self.nodes[n].get("is_anchor")
            if is_anchor:
                return n

    def node_sparql_variable(self, node):
        return self.nodes[node]["sparql_var"]

    def node_instance_iri(self, node):
        return self.nodes[node].get("instance_iri")

    def print_edges(self):
        for u, v, relation in self.edges(data="relation"):
            print("{} {} {}".format(u, relation, v))

    def generate_sparql_representation(self):
        conditions = []
        for n in self:
            type_declaration = "?{} rdf:type {}".format(self.node_sparql_variable(n), self.node_class(n))
            conditions.append(type_declaration)
        # TODO: Check that order of assertion triples in SPARQL doesn't affect result
        for u, v, relation in self.edges(data="relation"):
            assert_declaration = "?{} {} ?{}".format(self.node_sparql_variable(u), relation, self.node_sparql_variable(v))
            conditions.append(assert_declaration)

        return " .\n".join(conditions)

    def find_matches_in_model(self, model):
        sparql_wrapper = RdflibSparqlWrapper()
        query = """
                    select *
                    WHERE {{
                        {}
                    }}
                """.format(self.generate_sparql_representation())
        return sparql_wrapper.run_query(model.graph, query)

    def print_matches_in_model(self, model):
        # For debugging
        response = self.find_matches_in_model(model)

        if len(response) > 0:
            count = 1
            for res in response:
                print("Result {}:".format(count))
                for n in self:
                    instance_iri = res[self.node_sparql_variable(n)]
                    print("{} - {}".format(n, instance_iri))
                count += 1

    def write_to_model(self, model, evidences, reuse_existing=False):
        # Handles reusing found subgraph IRIs as well as inserting all new individuals
        axiom_ids = []
        exact_match = None
        response = []
        if reuse_existing:
            response = self.find_matches_in_model(model)
        if len(response) > 0:
            for res in response:
                # TODO: First check that these are "exact" subgraph matches, meaning match result isn't subgraph of other annotation.
                exact_match = res
                break
        if exact_match:
            # Update rdflib.Graph - just add evidence to all edges/axioms
            # Collect axiom_ids to add evidence to
            for u, v, relation in self.edges(data="relation"):
                subject_instance_iri = exact_match[self.node_sparql_variable(u)]
                object_instance_iri = exact_match[self.node_sparql_variable(v)]
                self.nodes[u]["instance_iri"] = subject_instance_iri
                self.nodes[v]["instance_iri"] = object_instance_iri
                relation_uri = expand_uri_wrapper(relation)
                axiom_ids.append(model.find_bnode((self.node_instance_iri(u),
                                                   URIRef(relation_uri),
                                                   self.node_instance_iri(v))))
        else:
            # Insert into rdflib.Graph - everything is new
            pattern = self.generate_sparql_representation()
            # No. Need to invent IRIs for individuals.
            # Can we reuse model.declare_individual, add_axiom stuff? Probably.
            for u, v, relation in self.edges(data="relation"):
                subject_instance_iri = self.node_instance_iri(u)
                if subject_instance_iri is None:
                    subject_instance_iri = model.declare_individual(self.node_class(u))
                    self.nodes[u]["instance_iri"] = subject_instance_iri
                object_instance_iri = self.node_instance_iri(v)
                if object_instance_iri is None:
                    object_instance_iri = model.declare_individual(self.node_class(v))
                    self.nodes[v]["instance_iri"] = object_instance_iri
                try:
                    relation_uri = expand_uri_wrapper(relation)
                except AttributeError as ex:
                    exception_message = "Unparseable relation: {relation} from triple {u} {relation} {v} in model {modeltitle}".format(
                        relation=relation, u=u, v=v, modeltitle=model.modeltitle)
                    logger.info(exception_message)
                    raise ModelRdfWriteException(exception_message)
                axiom_ids.append(model.add_axiom(model.writer.emit(subject_instance_iri,
                                                                   URIRef(relation_uri),
                                                                   object_instance_iri)))
        # Add the evidences to whatever axioms we got
        for axiom_id in axiom_ids:
            for evidence in evidences:
                model.add_evidence(axiom_id, evidence)


    @staticmethod
    def node_class(node):
        return "-".join(node.split("-")[0:-1])



import pytest

from ontobio import ontol
from ontobio import ontol_factory

def test_missing_node_is_none():
    ontology = ontol_factory.OntologyFactory().create("tests/resources/goslim_generic.json")

    assert ontology.node("GO:0") == None

def test_merge_copies_logical_definitions():
    pombe_ontology = ontol_factory.OntologyFactory().create("tests/resources/goslim_pombe.json")
    assert len(pombe_ontology.all_logical_definitions) == 0
    nucleus_ontology = ontol_factory.OntologyFactory().create("tests/resources/nucleus.json")
    assert len(nucleus_ontology.all_logical_definitions) == 2
    # Test logical definition copy in merge
    pombe_ontology.merge([nucleus_ontology])
    assert len(pombe_ontology.all_property_chain_axioms) == 2

def test_merge_copies_property_chain_axioms():
    nucleus_ontology = ontol_factory.OntologyFactory().create("tests/resources/nucleus.json")
    assert len(nucleus_ontology.all_property_chain_axioms) == 0
    goslim_ontology = ontol_factory.OntologyFactory().create("tests/resources/goslim_generic.json")
    assert len(goslim_ontology.all_property_chain_axioms) == 2
    # Test property chain axiom copy in merge
    nucleus_ontology.merge([goslim_ontology])
    assert len(nucleus_ontology.all_property_chain_axioms) == 2

def test_ontology_synonyms():
    ontology = ontol_factory.OntologyFactory().create("tests/resources/nucleus_new.json")
    syn = ontology.synonyms("GO:0005634")

    assert syn[0].__dict__ == ontol.Synonym("GO:0005634", val="cell nucleus", pred="hasExactSynonym", lextype=None,
                        xrefs=[], ontology=None, confidence=1.0, synonymType="http://purl.obolibrary.org/obo/go-test#systematic_synonym").__dict__



from os import path
import json
import logging
import requests
from typing import List
from ontobio.rdfgen.assoc_rdfgen import prefix_context
from ontobio.rdfgen.gocamgen.errors import ShexException
from ontobio.model.association import Curie
from prefixcommons.curie_util import expand_uri, contract_uri
from pyshexc.parser_impl import generate_shexj

logger = logging.getLogger(__name__)


def expand_uri_wrapper(id):
    uri = expand_uri(id, cmaps=[prefix_context])
    return uri


def contract_uri_wrapper(id):
    uri = contract_uri(id, cmaps=[prefix_context])
    return uri


def sort_terms_by_ontology_specificity(terms: List[Curie]):
    # Used primarily for sorting occurs_in annotation extensions
    # What's first? EMAPA or UBERON? Shouldn't matter for extensions since assertion
    # should chain occurs_in to both EMAPA and UBERON (They would be split into separate assertions or thrown out)
    ONTOLOGY_ORDER = {'SO': 1, 'GO': 2, 'CL': 3, 'WBbt': 4, 'EMAPA': 5, 'UBERON': 6}  # From most specific to most general
    terms.sort(key=lambda t: ONTOLOGY_ORDER[t.namespace])

    return terms


class ShexHelper:
    def __init__(self):
        self.shapes = None

    def load_shapes(self):
        self.shapes = {}
        shapes_to_load = ["ProteinContainingComplex", "MolecularFunction", "CellularComponent", "BiologicalProcess",
                          "AnatomicalEntity", "InformationBiomacromolecule"]
        
        shex_url = "https://raw.githubusercontent.com/geneontology/go-shapes/master/shapes/go-cam-shapes.shex"
        shex_response = requests.get(shex_url)
        shex_raw = shex_response.text
        shex_json_str = generate_shexj.parse(shex_raw)._as_json_dumps()
        full_shex_ds = json.loads(shex_json_str)
        for shape in full_shex_ds["shapes"]:
            shape_name = path.basename(shape["id"])
            if shape_name in shapes_to_load:
                self.shapes[shape_name] = {}
                shexps = shape.get('shapeExprs')
                if shexps is None:
                    shexps = [shape]
                for shexp in shexps:
                    if isinstance(shexp, dict) and 'expression' in shexp:
                        for exp in shexp['expression']['expressions']:
                            if exp['type'] == 'TripleConstraint':
                                predicate = contract_uri_wrapper(exp['predicate'])[0]
                                self.shapes[shape_name][predicate] = []
                                values = exp['valueExpr']
                                if isinstance(values, dict):
                                    values = values['shapeExprs']
                                else:
                                    values = [values]
                                for v in values:
                                    # path.basename(v) - Gets the Shape name minus the URL prefix
                                    # E.g. -
                                    self.shapes[shape_name][predicate].append(path.basename(v))
                del self.shapes[shape_name]["rdf:type"]
        # TODO: Get this into ShEx spec or delete this when we decide against it
        self.shapes["CellularComponent"]["BFO:0000050"].append("CellularComponent")  # CC-part_of->CC

    # Try to minimize replication of ShEx structure/logic since it ideally should be centralized elsewhere
    def shape_from_class(self, class_term, go_aspector):
        complex_term = "GO:0032991"
        # TODO: Can we pull these root terms for shape from spec?
        shape_map = {
            complex_term: "ProteinContainingComplex",
            "GO:0003674": "MolecularFunction",
            "GO:0005575": "CellularComponent",
            "GO:0008150": "BiologicalProcess",
            "CARO:0000000": "AnatomicalEntity",
        }
        aspect = go_aspector.go_aspect(class_term)
        if aspect == "C":
            # Oh great, now we gotta look further into the ontology
            # TODO: ensure reflexive=True in all ancestors calls in ontobio.util.go_utils.get_ancestors_through_subont()
            if complex_term in go_aspector.get_isa_closure(class_term) + [class_term]:
                return shape_map[complex_term]
            else:
                return shape_map["GO:0005575"]
        elif aspect == "F":
            return shape_map["GO:0003674"]
        elif aspect == "P":
            return shape_map["GO:0008150"]
        else:
            return shape_map["CARO:0000000"]

    def relation_lookup(self, subj_shape, obj_shape):
        # Well, crud. Guess we're gunna do this:
        if self.shapes is None:
            self.load_shapes()
        # This will screw up the CC->AnatomicalEntity relations
        # <CellularComponent> @<GoCamEntity> AND EXTRA a {
        #   a ( @<CellularComponentClass> OR @<NegatedCellularComponentClass> ) {1};
        #   part_of: @<AnatomicalEntity> {0,1};
        #   adjacent_to: @<AnatomicalEntity> *;
        #   overlaps: @<AnatomicalEntity> *;
        obj_shapes = [obj_shape]
        if obj_shape == "CellularComponent":
            # Add AnatomicalEntity for CC. Should first look for specific CC matches then check for AnatomicalEntity.
            obj_shapes.append("AnatomicalEntity")
        for relation, allowable_obj_shapes in self.shapes[subj_shape].items():
            for obs in obj_shapes:
                if obs in allowable_obj_shapes:
                    return relation
        # Oh no, we couldn't find a relation. This is a no-go for this model.
        shex_error_message = "No relation found in ShEx for {}->{}".format(subj_shape, obj_shape)
        logger.warning(shex_error_message)
        # Throw Exception and except-skip in model builder
        raise ShexException(shex_error_message)



from ontobio.ontol_factory import OntologyFactory
import logging

HAS_PART = 'BFO:0000051'
PART_OF = 'BFO:0000050'

QUALITY = 'PATO:0000001'
PLOIDY = 'PATO:0001374'
EUPLOID = 'PATO:0001393'
SHAPE = 'PATO:0000052'
Y_SHAPED = 'PATO:0001201'
PENTAPLOID = 'PATO:0001383'
SWOLLEN = 'PATO:0001851'
DILATED = 'PATO:0001571'
INCREASED_SIZE = 'PATO:0000586'
PROTRUDING = 'PATO:0001598'
MORPHOLOGY = 'PATO:0000051'
ABSENT = 'PATO:0000462'
CONICAL = 'PATO:0002021'

def test_remote_sparql():
    """
    Load ontology from remote SPARQL endpoint
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont = factory.create('scigraph:ontology')

    ploidy = ont.node(PLOIDY)
    print("PLOIDY: {}".format(ploidy))
    assert ont.label(PLOIDY) == 'ploidy'

    # exact match
    search_results = ont.search('shape')
    print("SEARCH (exact): {}".format(search_results))
    assert [SHAPE] == search_results

    # implicit regexp
    search_results = ont.search('%lantern%')
    print("SEARCH (re, implicit): {}".format(search_results))
    assert 'UBERON:0008253' in search_results
    
    # syns
    syn = 'cone-shaped'
    search_results = ont.search(syn, synonyms=False)
    print("SEARCH (no syns): {}".format(search_results))
    assert [] == search_results
    #search_results = ont.search(syn, synonyms=True)
    #print("SEARCH (with syns): {}".format(search_results))
    #assert [CONICAL] == search_results
    

    ancs = ont.ancestors(PLOIDY, relations=['subClassOf'])
    print("ANCS ploidy (subClassOf): {}".format(ancs))
    assert QUALITY in ancs
    assert PENTAPLOID not in ancs

    # this is a non-use case
    ancs = ont.ancestors(SWOLLEN, relations=[HAS_PART])
    print("ANCS swollen (has_part): {}".format(ancs))
    assert INCREASED_SIZE in ancs
    assert PROTRUDING in ancs
    assert len(ancs) == 2

    ancs = ont.ancestors(SWOLLEN, relations=['subClassOf'], reflexive=False)
    print("ANCS swollen (has_part): {}".format(ancs))
    assert MORPHOLOGY in ancs
    assert QUALITY in ancs
    assert PROTRUDING not in ancs
    
    #decs = ont.descendants(PLOIDY)
    #print("DECS ploidy (all): {}".format(decs))
    #assert QUALITY not in decs
    #assert EUPLOID in decs
    #assert PENTAPLOID in decs

    # this is a non-use case
    ancs = ont.descendants(INCREASED_SIZE, relations=[HAS_PART])
    print("ANCS increased size (has part): {}".format(ancs))
    assert SWOLLEN in ancs
    assert len(ancs) == 1

    #subsets = ont.subsets()
    #print("SUBSETS: {}".format(subsets))

    #slim = ont.extract_subset('absent_slim')
    #print("SLIM: {}".format(slim))
    #assert ABSENT in slim
    #assert QUALITY not in slim

#def test_subontology():
def todo_subontology():
    """
    subontology
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont = factory.create('scigraph:ontology')
    print("ONT NODES: {}".format(ont.nodes()))

    # TODO: implement this for SciGraph:
    subont = ont.subontology(relations=['subClassOf'])
    PERM = 'GO:1990578'
    print("NODES: {}".format(subont.nodes()))
    ancs = subont.ancestors(PERM, reflexive=True)
    print(str(ancs))
    for a in ancs:
        print(" ANC: {} '{}'".format(a,subont.label(a)))
    assert len(ancs) > 0
    from ontobio.io.ontol_renderers import GraphRenderer
    w = GraphRenderer.create('tree')
    w.write_subgraph(ont, ancs)

    



from ontobio.sparql.sparql_ontol_utils import SEPARATOR

def correctGOID(self, goid):
    return goid.replace(":", "_")

def correctGOCAM(self, gocam):
    if gocam.find('/') != -1:
        gocam = gocam[gocam:gocam.rfind('/') + 1]
    if not gocam.startsWith('gomodel'):
        gocam = 'gomodel:' + gocam.strip()
    return gocam

def correctGOCAMs(self, gocams):
    list = []
    for gocam in gocams:
        list.append(correctGOCAM(self, gocam))
    return list


def goSummary(self, goid):
    goid = correctGOID(self, goid)
    return """
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX definition: <http://purl.obolibrary.org/obo/IAO_0000115>
    PREFIX obo: <http://www.geneontology.org/formats/oboInOwl#>

    SELECT ?goid ?label ?definition ?comment ?creation_date		(GROUP_CONCAT(distinct ?synonym;separator='""" + SEPARATOR + """') as ?synonyms)
                                                                (GROUP_CONCAT(distinct ?relatedSynonym;separator='""" + SEPARATOR + """') as ?relatedSynonyms)
                                                                (GROUP_CONCAT(distinct ?alternativeId;separator='""" + SEPARATOR + """') as ?alternativeIds)
                                                                (GROUP_CONCAT(distinct ?xref;separator='""" + SEPARATOR + """') as ?xrefs)
                                                                (GROUP_CONCAT(distinct ?subset;separator='""" + SEPARATOR + """') as ?subsets)

    WHERE {
        BIND(<http://purl.obolibrary.org/obo/""" + goid + """> as ?goid) .
        optional { ?goid rdfs:label ?label } .
        optional { ?goid definition: ?definition } .
        optional { ?goid rdfs:comment ?comment } .
        optional { ?goid obo:creation_date ?creation_date } .
        optional { ?goid obo:hasAlternativeId ?alternativeId } .
        optional { ?goid obo:hasRelatedSynonym ?relatedSynonym } .
        optional { ?goid obo:hasExactSynonym ?synonym } .
        optional { ?goid obo:hasDbXref ?xref } .
        optional { ?goid obo:inSubset ?subset } .
    }
    GROUP BY ?goid ?label ?definition ?comment ?creation_date
    """

def goHierarchy(self, goid):
    goid = correctGOID(self, goid)
    return """   
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX definition: <http://purl.obolibrary.org/obo/IAO_0000115>
    SELECT ?hierarchy ?GO ?label WHERE {
        BIND(<http://purl.obolibrary.org/obo/""" + goid + """> as ?goquery)
        {
            {
                ?goquery rdfs:subClassOf+ ?GO .
                ?GO rdfs:label ?label .
                FILTER (LANG(?label) != "en")    
                BIND("parent" as ?hierarchy)
            }
            UNION
            {
                ?GO rdfs:subClassOf* ?goquery .
                ?GO rdfs:label ?label .    		
                FILTER (LANG(?label) != "en")    
                BIND(IF(?goquery = ?GO, "query", "child") as ?hierarchy) .
            }
        }
    }  
    """


def gocams(self, states, start = None, rows = None):
    statesList = ""

    for docState in states:
        statesList += "\"" + docState.value + "\"^^xsd:string "

    query = """
    PREFIX metago: <http://model.geneontology.org/>
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX obo: <http://www.geneontology.org/formats/oboInOwl#>
    PREFIX providedBy: <http://purl.org/pav/providedBy>

    SELECT  ?gocam ?date ?state ?title (GROUP_CONCAT(?orcid;separator='""" + SEPARATOR + """') AS ?orcids) 
                                (GROUP_CONCAT(?name;separator='""" + SEPARATOR + """') AS ?names)
                                (GROUP_CONCAT(distinct ?providedBy;separator='""" + SEPARATOR + """') AS ?groupids) 
                                (GROUP_CONCAT(distinct ?providedByLabel;separator='""" + SEPARATOR + """') AS ?groupnames) 
    
    WHERE 
    {
        {
            VALUES ?state { """ + statesList + """ }
            GRAPH ?gocam {            
                ?gocam metago:graphType metago:noctuaCam .
                ?gocam <http://geneontology.org/lego/modelstate> ?state .
                ?gocam dc:title ?title ;
                     dc:date ?date ;
                     dc:contributor ?orcid .
                BIND( IRI(?orcid) AS ?orcidIRI ).
            }
     
            optional { ?orcidIRI rdfs:label ?name }
            BIND(IF(bound(?name), ?name, ?orcid) as ?name) .
        }
      
        optional {
            GRAPH ?gocam {
                ?gocam providedBy: ?providedBy .
                BIND( IRI(?providedBy) AS ?providedByIRI ).
            }
            ?providedByIRI rdfs:label ?providedByLabel .
        }   
      
    }
    GROUP BY ?gocam ?date ?title ?state
    ORDER BY DESC(?date)
    """
    if start is not None and rows is not None:
        query += """LIMIT """ + str(rows) + """
                    OFFSET """ + str(start)
    return query


def gocamsWithGO(self, goid):
    goid = correctGOID(self, goid)
    return """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX metago: <http://model.geneontology.org/>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>

    SELECT distinct ?gocam
    WHERE 
    {
        GRAPH ?gocam {
            ?gocam metago:graphType metago:noctuaCam .    
            ?entity rdf:type owl:NamedIndividual .
            ?entity rdf:type ?goid .
            FILTER(?goid = <http://purl.obolibrary.org/obo/""" + goid + """>)
        }
    }
    """

def gocamsWithPMID(self, pmid):
    return """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX metago: <http://model.geneontology.org/>

    SELECT distinct ?gocam
    WHERE 
    {
        GRAPH ?gocam {
            ?gocam metago:graphType metago:noctuaCam .    	
            ?s dc:source ?source .
            BIND(REPLACE(?source, " ", "") AS ?source) .
            FILTER((CONTAINS(?source, """ + pmid + """)))
        }           
    }
    """

def gocamsWithGP(self, gpid):
    return """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX metago: <http://model.geneontology.org/>
    PREFIX enabled_by: <http://purl.obolibrary.org/obo/RO_0002333>
    
    SELECT distinct ?gocam
    WHERE 
    {  
      GRAPH ?gocam {
        ?gocam metago:graphType metago:noctuaCam .    
        ?s enabled_by: ?gpnode .    
        ?gpnode rdf:type ?identifier .
        FILTER(?identifier = <""" + gpid + """>) .         
      } 
    }
    """







def gocamsGO(self, gocams):
    gocams = correctGOCAMs(self, gocams)

    return """
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX metago: <http://model.geneontology.org/>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX definition: <http://purl.obolibrary.org/obo/IAO_0000115>
    PREFIX BP: <http://purl.obolibrary.org/obo/GO_0008150>
    PREFIX MF: <http://purl.obolibrary.org/obo/GO_0003674>
    PREFIX CC: <http://purl.obolibrary.org/obo/GO_0005575>

    SELECT distinct ?gocam ?goclasses ?goids ?gonames ?definitions
    WHERE 
    {
        VALUES ?gocam { """ + gocams + """}
#		    VALUES ?gocam { <http://model.geneontology.org/5a7e68a100001298> <http://model.geneontology.org/5a7e68a100001201> <http://model.geneontology.org/5a7e68a100001125> <http://model.geneontology.org/5a7e68a100000655>}

        GRAPH ?gocam {
            ?entity rdf:type owl:NamedIndividual .
            ?entity rdf:type ?goids
        }

        VALUES ?goclasses { BP: MF: CC:  } . 
        ?goids rdfs:subClassOf+ ?goclasses .
        ?goids rdfs:label ?gonames .
        ?goids definition: ?definitions .
    }
    ORDER BY DESC(?gocam)
    """




def userMeta(self, orcid):
    return """
    PREFIX metago: <http://model.geneontology.org/>
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX vcard: <http://www.w3.org/2006/vcard/ns#>
    PREFIX has_affiliation: <http://purl.obolibrary.org/obo/ERO_0000066> 
    PREFIX enabled_by: <http://purl.obolibrary.org/obo/RO_0002333>
    PREFIX obo: <http://www.geneontology.org/formats/oboInOwl#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX BP: <http://purl.obolibrary.org/obo/GO_0008150>
    PREFIX MF: <http://purl.obolibrary.org/obo/GO_0003674>
    PREFIX CC: <http://purl.obolibrary.org/obo/GO_0005575>

    SELECT  ?name  			(GROUP_CONCAT(distinct ?organization;separator='""" + SEPARATOR + """') AS ?organizations) 
                            (GROUP_CONCAT(distinct ?affiliationIRI;separator='""" + SEPARATOR + """') AS ?affiliationsIRI) 
                            (GROUP_CONCAT(distinct ?affiliation;separator='""" + SEPARATOR + """') AS ?affiliations) 
                            (GROUP_CONCAT(distinct ?gocam;separator='""" + SEPARATOR + """') as ?gocams)
                            (GROUP_CONCAT(distinct ?date;separator='""" + SEPARATOR + """') as ?gocamsDate)
                            (GROUP_CONCAT(distinct ?title;separator='""" + SEPARATOR + """') as ?gocamsTitle)
                            (GROUP_CONCAT(distinct ?goid;separator='""" + SEPARATOR + """') as ?bpids)
                            (GROUP_CONCAT(distinct ?goname;separator='""" + SEPARATOR + """') as ?bpnames)
                            (GROUP_CONCAT(distinct ?gpid;separator='""" + SEPARATOR + """') as ?gpids)
                            (GROUP_CONCAT(distinct ?gpname;separator='""" + SEPARATOR + """') as ?gpnames)
    WHERE 
    {
        #BIND(""" + orcid + """ as ?orcid) .
        #BIND("SynGO:SynGO-pim"^^xsd:string as ?orcid) .
        BIND("http://orcid.org/0000-0001-7476-6306"^^xsd:string as ?orcid)
        #BIND("http://orcid.org/0000-0003-1074-8103"^^xsd:string as ?orcid) .

        BIND(IRI(?orcid) as ?orcidIRI) .


        # Getting some information on the model
        GRAPH ?gocam 
        {
            ?gocam 	metago:graphType metago:noctuaCam ;
                    dc:date ?date ;
                    dc:title ?title ;
                    dc:contributor ?orcid .

            ?entity rdf:type owl:NamedIndividual .
            ?entity rdf:type ?goid .

            ?s enabled_by: ?gpentity .    
            ?gpentity rdf:type ?gpid .
            FILTER(?gpid != owl:NamedIndividual) .
        }


        VALUES ?GO_class { BP: } . 
        # rdf:type faster then subClassOf+ but require filter 			
        # ?goid rdfs:subClassOf+ ?GO_class .
        ?entity rdf:type ?GO_class .

        # Filtering out the root BP, MF & CC terms
        filter(?goid != MF: )
        filter(?goid != BP: )
        filter(?goid != CC: )

        ?goid rdfs:label ?goname .

        # Getting some information on the contributor
        optional { ?orcidIRI rdfs:label ?name } .
        BIND(IF(bound(?name), ?name, ?orcid) as ?name) .
        optional { ?orcidIRI vcard:organization-name ?organization } .
        optional { 
            ?orcidIRI has_affiliation: ?affiliationIRI .
            ?affiliationIRI rdfs:label ?affiliation
        } .

        # crash the query for SYNGO user "http://orcid.org/0000-0002-1190-4481"^^xsd:string  
        optional {
        ?gpid rdfs:label ?gpname .
        }
        BIND(IF(bound(?gpname), ?gpname, ?gpid) as ?gpname)

    }
    GROUP BY ?name    
    """


def goSubsets(self, goid):
    goid = correctGOID(self, goid)
    return """
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX obo: <http://www.geneontology.org/formats/oboInOwl#>

    SELECT ?label ?subset

    WHERE {
        BIND(<http://purl.obolibrary.org/obo/""" + goid + """> as ?goid) .
        optional { ?goid obo:inSubset ?subset .
                   ?subset rdfs:comment ?label } .
    }
    """

def subsetIdIRI(self, subsetid):
    return "http://purl.obolibrary.org/obo/go#" + subsetid

def subset(self, subsetid):
    subseturi = subsetIdIRI(self, subsetid)
    return """
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX obo: <http://www.geneontology.org/formats/oboInOwl#>

    SELECT ?goid ?label
    WHERE {
        BIND(<""" + subseturi + """> as ?subset) .  
	  	?goid obo:inSubset ?subset .
  		?goid rdfs:label ?label .
		FILTER (lang(?label) != "en")
    }
    """


"""
Wrap WD sparql
"""

from SPARQLWrapper import SPARQLWrapper, JSON
from prefixcommons.curie_util import contract_uri, expand_uri
from functools import lru_cache
import networkx
import logging


# CACHE STRATEGY:
# by default, the cache is NOT persistent. Only single threaded clients should
# call a cached method with writecache=True.
# Note we are layering the in-memory cache over the persistent cache

cache = lru_cache(maxsize=None)

LIMIT = 200000

prefixmap = dict(
    HP = 'http://www.wikidata.org/prop/direct/P3841',
    ENVO = 'http://www.wikidata.org/prop/direct/P3859',
    DOID = 'http://www.wikidata.org/prop/direct/P699',
    CHEBI = 'http://www.wikidata.org/prop/direct/P683',
    UniProtKB = 'http://www.wikidata.org/prop/direct/P352',
    NCBIGene = 'http://www.wikidata.org/prop/direct/P351',
    IPR = 'http://www.wikidata.org/prop/direct/P2926',
    encodes = 'http://www.wikidata.org/prop/direct/P688',
    genetic_association = 'http://www.wikidata.org/prop/direct/P2293',
    treated_by_drug = 'http://www.wikidata.org/prop/direct/P2176')

qmap = dict(
    disease2protein = dict(chain=['genetic_association', 'encodes'], prefix='UniProtKB'),
    disease2drug = dict(chain=['treated_by_drug']),
    genetic_association = dict(chain=['genetic_association'], prefix='NCBIGene')
    )

logger = logging.getLogger(__name__)


class Properties:
    """
    Common SPARQL prefixes used by wikidata.

    Note we use the "trick" whereby an entire property URI can be encoded as a prefix.
    """
    def prefixes(self):
        return [attr for attr in dir(self) if not callable(getattr(self,attr)) and not attr.startswith("__")]
    def get_uri(self, pfx):
        return vars(PrefixMap).get(pfx)
    def gen_header(self):
        return "\n".join(["prefix {}: <{}>".format(attr,self.get_uri(attr)) for attr in self.prefixes()])

    DOID = 'http://www.wikidata.org/prop/direct/P699'
    CHEBI = 'http://www.wikidata.org/prop/direct/P683'

def run_sparql(q):
    # TODO: select endpoint based on ontology
    logger.info("Connecting to sparql endpoint...")
    sparql = SPARQLWrapper("http://query.wikidata.org/sparql")
    logger.info("Made wrapper: {}".format(sparql))
    # TODO: iterate over large sets?
    full_q = q + ' LIMIT ' + str(LIMIT)
    sparql.setQuery(q)
    sparql.setReturnFormat(JSON)
    logger.info("Query: {}".format(q))
    results = sparql.query().convert()
    bindings = results['results']['bindings']
    logger.info("Rows: {}".format(len(bindings)))
    for r in bindings:
        curiefy(r)
    return bindings

def curiefy(r):
    for (k,v) in r.items():
        if v['type'] == 'uri':
            curies = contract_uri(v['value'])
            if len(curies)>0:
                r[k]['value'] = curies[0]

def ensure_prefixed(x, prefix):
    if not x.startswith(prefix):
        x = prefix + ':' + x
    return x
                
def fetchall_xrefs(prefix):
    m = {}
    pairs = fetchall_triples_xrefs(prefix)
    if len(pairs) == LIMIT:
        logger.error("Too many: {} {}".format(LIMIT, pairs[:10]))
        return None
    for (c,x) in pairs:
        if not x.startswith(prefix):
            x = prefix + ':' + x
        if x not in m:
            m[x] = []
        m[x].append(c)
    return m

def fetchall_triples_xrefs(prefix):
    """
    fetch all xrefs for a prefix, e.g. CHEBI
    """
    logger.info("fetching xrefs for: "+prefix)
    query = """
    SELECT * WHERE {{
    ?c <{p}> ?x
    }}
    """.format(p=prefixmap[prefix])
    bindings = run_sparql(query)
    rows = [(r['c']['value'], r['x']['value']) for r in bindings]
    return rows

def map_id(id, prefix):
    m = fetchall_xrefs(prefix)
    if id in m:
        return m[id]
    else:
        return [id]
    

def fetchall_sp(s,p):
    """
    fetch all triples for a property
    """
    query = """
    SELECT * WHERE {{
    <{s}> <{p}> ?x
    }}
    """.format(s=s, p=prefixmap[p])
    bindings = run_sparql(query)
    rows = [r['x']['value'] for r in bindings]
    return rows

def canned_query(n,s):
    """
    uses canned query
    """
    qobj = qmap[n]
    chain = qobj['chain']
    prefix = qobj['prefix']
    whr = chain2where(chain + [prefix], 'x')
    query = """
    SELECT * WHERE {{
    <{s}> {w}
    }}
    """.format(s=s, w=whr)
    bindings = run_sparql(query)
    rows = [ensure_prefixed(r['x']['value'], prefix) for r in bindings]
    return rows


def chain2where(chain, v):
    [head] = chain[0:1]
    p = prefixmap[head]
    if len(chain) == 1:
        return "<{}> ?{}".format(p, v)
    else:
        tail = chain[1:]
        return "<{}> [ {} ]".format(p, chain2where(tail, v))

# TODO: move to factory
#def make_aset(



from __future__ import absolute_import


from .ontol_renderers import GraphRenderer

import re
parser_version_regex = re.compile(r"!([\w]+)-version:[\s]*([\d]+\.[\d]+(\.[\d]+)?)")

from .assocwriter import GafWriter, GpadWriter



import typing
import logging
from dataclasses import dataclass, field
from typing import List, Dict, Callable

from ontobio.model import association
from ontobio.model.association import GoAssociation, Subject, Curie, Union, Optional
from ontobio.io import parser_version_regex
from ontobio.io import assocparser, gafparser, gpadparser, entityparser

logger = logging.getLogger(__name__)

@dataclass
class BioEntities:
    entities: Dict[Curie, Subject]

    def merge(self, other):
        """
        Merge another BioEntity set into this one. The `other` set will
        override any collisions in this BioEntities
        """
        self.entities.update(other.entities)
        return self

    def get(self, entity_id: Curie) -> Optional[Subject]:
        """
        Using an entity ID Curie, attempt to obtain the BioEntity `Subject` with
        this ID. If not present, this returns `None`.
        """
        return self.entities.get(entity_id, None)

    @classmethod
    def load_from_file(BioEntities, path: str):
        entities = dict() # type: Dict[Curie, Subject]
        print("loading from {}".format(path))
        try:
            gpi_parser = entityparser.GpiParser()
            with open(path) as gpi:
                for line in gpi:

                    ents = gpi_parser.line_as_entity_subject(line)
                    if ents is None:
                        continue

                    for entity in ents:
                        entity_id = entity.id
                        entities[entity_id] = entity

        except Exception as e:
            logger.error("Failed to read GPI file: {}".format(str(e)))

        return BioEntities(entities)

@dataclass
class GoAssociations:
    associations: List[GoAssociation]

@dataclass
class AssociationCollection:
    headers: List[str]
    associations: GoAssociations
    entities: BioEntities
    report: assocparser.Report

    @classmethod
    def initial(cls):  # type AssociationCollection
        return AssociationCollection([], GoAssociations([]), BioEntities(dict()), assocparser.Report())


def create_parser_from_header(line: str, config: assocparser.AssocParserConfig, group="unknown", dataset="unknown", bio_entities=None) -> Optional[assocparser.AssocParser]:
    parser = None
    parsed_version = parser_version_regex.findall(line)
    if len(parsed_version) == 1:
        filetype, version, _ = parsed_version[0]
        if filetype in ["gpad", "gpa"]:
            parser = gpadparser.GpadParser(config=config, bio_entities=bio_entities, group=group, dataset=dataset)
            if version in ["1.2", "2.0"]:
                parser.version = version
        elif filetype == "gaf":
            parser = gafparser.GafParser(config=config, bio_entities=bio_entities, group=group, dataset=dataset)
            if version in ["2.1", "2.2"]:
                parser.version = version

    return parser


def construct_collection(annotation_path: Optional[str], gpi_paths: List[str], config: assocparser.AssocParserConfig) -> AssociationCollection:
    entities = BioEntities(dict())
    for gpi in gpi_paths:
        entities.merge(BioEntities.load_from_file(gpi))

    annotations = []
    headers = []
    report = assocparser.Report()
    if annotation_path:
        parser = GeneralAssocParser(config, "unknown", bio_entities=entities)
        try:
            annotations = parser.parse(annotation_path)
        except NoVersionInFile as e:
            parser.report = e.report
            annotations = []

        headers = parser.headers
        report = parser.report

    return AssociationCollection(headers, GoAssociations(annotations), entities, report)


@dataclass
class NoVersionInFile(Exception):
    message: str
    report: assocparser.Report

    def __str__(self):
        return "NoVersionInFile: {}".format(self.message)

@dataclass
class GeneralAssocParser(assocparser.AssocParser):
    config: assocparser.AssocParserConfig
    group: str
    headers: List[str] = field(default_factory=list)
    report: assocparser.Report = field(default_factory=lambda: assocparser.Report())
    bio_entities: BioEntities = field(default_factory=lambda: BioEntities(dict()))
    annotation_parser: Optional[assocparser.AssocParser] = None

    def parse_line(self, line):
        parsed = super().validate_line(line)
        if parsed:
            return parsed

        if self.is_header(line):
            if self.annotation_parser is None:
                # We're still looking for a version identifier
                parser = create_parser_from_header(line, self.config, bio_entities=self.bio_entities)
                if parser is not None:
                    self.annotation_parser = parser

            self.headers.append(line)
            return assocparser.ParseResult(line, [], skipped=False)

        # At this point, we are not going through headers, and so we should have selected a parser
        if self.annotation_parser is None:
            logger.error("File is bad! We need to bail here!")
            self.report.error("", "Invalid Annotation File", "", "File header is missing file and version", rule=1)
            raise NoVersionInFile("File has no version info", self.report)

        # Just hand off parse responsibility to underlying `annotation_parser`
        return self.annotation_parser.parse_line(line)



from ontobio.ontol_factory import OntologyFactory
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.assocmodel import AssociationSet
import logging
import random

CVP = 'MP:0004084' # cardiomyopathy
MUS = 'NCBITaxon:10090'

def test_construct():
    """
    enrichment test

    build a gene set from MP term for cardiomyopathy;
    test for enrichment against GO
    """
    ofactory = OntologyFactory()
    afactory = AssociationSetFactory()
    logging.info("Creating mp handle")
    mp = ofactory.create('obo:mp')
    logging.info("Getting pheno assocs")
    aset_phen = afactory.create(ontology=mp,
                                subject_category='gene',
                                object_category='phenotype',
                                taxon=MUS)
    
    logging.info("Creating go handle")
    ont = ofactory.create('go')
    logging.info("Getting go assocs")
    aset = afactory.create(ontology=ont,
                           subject_category='gene',
                           object_category='function',
                           taxon=MUS)

    logging.info("Getting sample")
    sample = aset_phen.query([CVP],[])
    logging.info("sample = {}".format(len(sample)))

    rs = aset.enrichment_test(sample, threshold=1e-2, labels=True, direction='less')
    for r in rs:
        print("UNDER: "+str(r))

    
    rs = aset.enrichment_test(sample, threshold=0.05, labels=True)
    for r in rs:
        print(str(r))


#test_construct()



import json
from typing import Dict, NewType, List

from ontobio.model.association import GoAssociation

class Entity(dict):

    def __init__(self, d):
        super(Entity, self).__init__(d)

    def __hash__(self):
        d = json.dumps(self, sort_keys=True)
        return hash(d)


class GafGpiBridge(object):

    def __init__(self):
        self.cache = []

    def convert_association(self, association) -> Entity:
        """
        'id' is already `join`ed in both the Association and the Entity,
        so we don't have to worry about what that looks like. We assume
        it's correct.
        """
        if isinstance(association, GoAssociation):
            # print(json.dumps(association, indent=4))
            gpi_obj = {
                'id': str(association.subject.id),
                'label': association.subject.label,  # db_object_symbol,
                'full_name': association.subject.fullname,  # db_object_name,
                'synonyms': association.subject.synonyms,
                'type': [str(t) for t in association.subject.type], #db_object_type,
                'parents': "", # GAF does not have this field, but it's optional in GPI
                'xrefs': "", # GAF does not have this field, but it's optional in GPI
                'taxon': {
                    'id': str(association.subject.taxon)
                }
            }
            return Entity(gpi_obj)

        return None

    def entities(self) -> List[Entity]:
        return list(self.cache)
